/*
 NSF.c
  Nintendo Sound File playback and view

 * MacFCEU 0.8a3
 *
 * Copyright notice for this file:
 *  nsf.c (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>

#include "driver.h"
#include "x6502.h"
#include "FCEU.h"
#include "fsound.h"
#include "NSF.h"
#include "general.h"
#include "memory.h"
#include "file.h"
#include "FDS.h"
#include "cart.h"
#include "input.h"
#include "video.h"
#include "PPU.h"
#include "boards/MMC5.h"
#include "boards/Namcot.h"
#include "iNES.h"	//for prototypes only

#include "utility/ftrig.h"

#define cosmul(r,t)	(ifract30mul(r+r,fastcos(t))/2)
#define sinmul(r,t)	(ifract30mul(r+r,fastsin(t))/2)

#ifdef env68k
 #pragma segment shell
#endif

/* ---- the MacFCEU NSF ROM */

/* revised: 2013.05.03 */
const uint8	NSFROM[128] =
{ //NMI
/*00*/	0x8D,0x04,0x3F,	/* STA	$3F04	; Stop play-routine NMIs.	*/
/*03*/	0x48,		/* PHA		*/
/*04*/	0xAD,0x00,0x3F,	/* LDA	$3F00	; Reload?  If not, play, RTI.	*/
/*07*/	0x30,0x11,	/* BMI	$3E1A	*/
/*09*/	0x8A,		/* TXA		; Save registers, and play.	*/
/*0A*/	0x48,		/* PHA		*/
/*0B*/	0x98,		/* TYA		*/
/*0C*/	0x48,		/* PHA		*/
/*0D*/	0xA9,0x00,	/* LDA	#$00	; Zero out registers.		*/
/*0F*/	0xAA,		/* TAX		*/
/*10*/	0xA8,		/* TAY		*/
/*11*/	0x20,0x7C,0x3E,	/* JSR	$3E7C	; JSR to play routine JMP.	*/
/*14*/	0x68,		/* PLA		*/
/*15*/	0xA8,		/* TAY		*/
/*16*/	0x68,		/* PLA		*/
/*17*/	0xAA,		/* TAX		*/
/*18*/	0x68,		/* PLA		*/
/*19*/	0x40,		/* RTI		*/
/*1A*/	0xA2,0xFF,	/* LDX	#$FF	; Initialize the stack pointer. */
/*1C*/	0x9A,		/* TXS		*/
/*1D*/	0xA2,0x07,	/* LDX	#$07	; Clear RAM.			*/
/*1F*/	0xBD,0x70,0x3F,	/* LDA	$3F70,X	; BankSwitch			*/
/*22*/	0x9D,0x08,0x3F,	/* STA	$3F08,X	; RAM page erase		*/
/*25*/	0x9D,0xF8,0x5F,	/* STA	$5FF8,X	; BANKSET			*/
/*28*/	0xE0,0x06,	/* CPX	#$06	*/
/*2A*/	0x30,0x06,	/* BMI	$3E32	*/
/*2C*/	0x9D,0x00,0x3F,	/* STA	$3F00,X	; ExRAM side erase		*/
/*2F*/	0x9D,0xF0,0x5F,	/* STA	$5FF0,X	; FDS BANKSET			*/
/*32*/	0xCA,		/* DEX		*/
/*33*/	0x10,0xEA,	/* BPL	$3E1F	*/
/*35*/	0x20,0x4A,0x3E,	/* JSR	$3E4A	; JSR to init routine and JMP.	*/
/*38*/	0xA9,0x00,	/* LDA	#$00	; Zero out registers.		*/
/*3A*/	0xAA,		/* TAX		*/
/*3B*/	0xA8,		/* TAY		*/
/*3C*/	0x20,0x7C,0x3E,	/* JSR	$3E7C	; JSR to play routine JMP.	*/
/*3F*/	0x8D,0x05,0x3F,	/* STA	$3F05	; Start play routine NMIs.	*/
/*42*/	0x90,0xFE,	/* BCC	$3E42	; Loopie time.			*/
/*44*/	0x8D,0x03,0x3F,	/* STA	$3F03	; Init init NMIs.		*/
/*47*/	0x18,		/* CLC		*/
/*48*/	0x90,0xFE,	/* BCC	$3E48 	; Loopie time.			*/
/*4A*/	0xA2,0x00,	/* LDX	#$00	; Init sound channels.		*/
/*4C*/	0x8A,		/* TXA		*/
/*4D*/	0x8D,0x15,0x40,	/* STA	$4015	*/
/*50*/	0x9D,0x00,0x40,	/* STA	$4000,X	*/
/*53*/	0xE8,		/* INX		*/
/*54*/	0xE0,0x14,	/* CPX	#$14	*/
/*56*/	0xD0,0xF8,	/* BNE	$3E50	*/
/*58*/	0xAD,0x7B,0x3F,	/* LDA	$3F7B	; Check FDS			*/
/*5B*/	0x29,0x04,	/* AND	#$04	*/
/*5D*/	0xF0,0x0A,	/* BEQ	$3E69	*/
/*5F*/	0xA9,0x80,	/* LDA	#$80	*/
/*61*/	0x8D,0x89,0x40,	/* STA	$4089	*/
/*64*/	0xA9,0xE8,	/* LDA	#$E8	*/
/*66*/	0x8D,0x8A,0x40,	/* STA	$408A	*/
/*69*/	0xA9,0x0F,	/* LDA	#$0F	; Final audio init.		*/
/*6B*/	0x8D,0x15,0x40,	/* STA	$4015	*/
/*6E*/	0xA9,0xC0,	/* LDA	#$C0	*/
/*70*/	0x8D,0x17,0x40,	/* STA	$4017	*/
/*73*/	0xAD,0x01,0x3F,	/* LDA	$3F01	*/
/*76*/	0xAE,0x7A,0x3F,	/* LDX	$3F80	; Load X with PAL.		*/
/*79*/	0x6C,0x0A,0x3F,	/* JMP	($3F0A)	*/
/*7C*/	0x6C,0x0C,0x3F,	/* JMP	($3F0C)	; Called by a JSR above.	*/
};

static DECLFR(NSF_ROMRead)
{
	return NSFROM[A&=0x7F];
}

/* ---- code */

#define MaxNamesTotal	(256*32)

static int GetIndName(const uint8 *l,
	uint8 *b,
	int i)
{
	const uint8	*s;

	while(1)
	{
		if ((s=memchr(l,0,MaxNamesTotal))==0)
			return 0;
		if (--i<0)
			break;
		l = s+1;
	}
	if ((i=s-l)>31)
		i = 31;
	memcpy(b,l,i);
	b[i] = 0;

	return 1;
}

static uint8	vismode = 1;

static uint8	*NSFDATA = 0;
static uint8	*ExWRAM = 0;

static uint8	*SongNames = 0;
static uint8	CurrentSongName[32], CaptionStr[32];

static int	NSFSize, NSFMaxBank;
static uint16	LoadAddr;

static NSF_HEADER	NSFHeader;
//cheap on memory? (yes); 2010.03.18
#define NSFNMIFlags	NSFHeader.reserved[0]
#define NSFCurrentSong	NSFHeader.reserved[1]
#define NSFSpecial	NSFHeader.reserved[2]
#define NSFBankSW	NSFHeader.reserved[3]
#define NSFSizeLo	NSFHeader.reserved[4]
#define NSFSizeHi	NSFHeader.reserved[5]
#define NSFLastSong	NSFHeader.reserved[6] //maximum
#define NSFMaxBankHi	NSFHeader.reserved[7]

#define NMIPendReload	1
#define NMIWrTrigger	2
#define NMIReset	4
#define NMISongReload	0x80

void NSFGI(int h)
{
	switch(h)
	{
	case GI_CLOSE:
		if (NSFDATA)
			{FCEU_free(NSFDATA); NSFDATA = 0;}
		if (ExWRAM)
			{FCEU_free(ExWRAM); ExWRAM = 0;}
		if (SongNames)
			{FCEU_free(SongNames); SongNames = 0;}
		//if (NSFHeader.SoundChip&1<<SCVRC6)
		//	NSFVRC6_Init();
		//if (NSFHeader.SoundChip&1<<SCVRC7)
		//	NSFVRC7_Init();
		//if (NSFHeader.SoundChip&1<<SCNFDS)
		//	FDSSoundReset();
		if (NSFHeader.SoundChip&1<<SCMMC5)
			GenMMC5Close();
		if (NSFHeader.SoundChip&1<<SCN106)
			GenN1xxClose();
		//if (NSFHeader.SoundChip&1<<SCSFME)
		//	NSFAY_Init();
		break;
	case GI_RESETM2:
	case GI_POWER:
		NSF_init();
	}
}

/* this expects NSFSize and NSFHeader.LoadAddressLow/High to already be set */
static int NSFLoadData(FCEUFILE *fp)
{
	if (NSFDATA)
		return 2;
	LoadAddr = NSFHeader.LoadAddressLow;
	LoadAddr|= NSFHeader.LoadAddressHigh<<8;

	if (LoadAddr<0x6000)
	{
		FCEUD_PrintError(" NSF Error: Invalid load address.");
		return -5;
	}
	NSFMaxBank = (NSFSize+(LoadAddr&0xFFF)+4095)/4096;
	NSFMaxBank = uppow2(NSFMaxBank);
	NSFMaxBankHi = NSFMaxBank>>8;
	if ((NSFDATA=(uint8*)FCEU_malloc(NSFMaxBank<<12))==0)
		return -2;
	return FCEU_fread(NSFDATA+(LoadAddr&0xFFF),NSFSize,1,fp);
}

// First 32KB is reserved for sound chip emulation in the iNES mapper code.

//static INLINE void BANKSET(uint32 A,
//	uint32 b)
#define BANKSET(A,b)	do{ uint32 _b=(b)&NSFMaxBank; \
	if (NSFHeader.SoundChip&1<<SCNFDS) \
		FCEU_memcpy(ExWRAM+((A)-0x6000),NSFDATA+(_b<<12),4096); \
	else \
		setprg4(A,_b); \
} while(0)

/* ---- NSFe */

typedef struct NSFeAtomHeader
{
	int32		size;
	uint32		tag;
} NSFeAtomHeader;

static int NSFeLoad(FCEUFILE *fp)
{
	uint8		buf[250];
	NSFeAtomHeader	*header = (NSFeAtomHeader*)buf;
	const char	*errs = "faile error in reading";
	int32		skip;

	LoadAddr = 0; /* not a good idea to assume a default here */
	memset(&NSFHeader,0,sizeof(NSFHeader));
	while(FCEU_fread(header,sizeof(NSFeAtomHeader),1,fp)==1)
	{
		header->size = le32_native(header->size);
		if ((skip=header->size)<0)
		{
			errs = "invalid atom data size";
			goto error;
		}
		switch(header->tag)
		{
		case 'INFO': /* the specifics escape me right now */
			memset(buf,0,12);
			/* the last two bytes are still unknown to me */
			if (FCEU_fread(buf,MIN(skip,12),1,fp)!=1)
				goto error;
			skip-= 12;
			memcpy(&NSFHeader.LoadAddressLow,buf,6);
			NSFHeader.VideoSystem = buf[6];
			NSFHeader.SoundChip = buf[7];
			NSFHeader.TotalSongs = buf[8];
			NSFHeader.StartingSong = buf[9];
			break;
		case 'BANK': /* loads almost directly */
			if ((skip-=8)<0)
				break;
			if (FCEU_fread(&NSFHeader.BankSwitch,8,1,fp)!=1)
				goto error;
			break;
		case 'time': /* time listings, by track */
			/* terminated w/ -1s but NotSoFatSo does not interpret */
			break;
		case 'fade': /* ditto; pre/post-track effect; not handled */
			break;
		case 'plst': /* song playlist (uint8s) */
			if (--skip<0)
				break;
			if (FCEU_fread(&NSFHeader.StartingSong,1,1,fp)!=1)
				goto error;
			if (!NSFHeader.StartingSong)
				NSFHeader.StartingSong = 1;
			break;
		case 'tlbl': /* C-str separated song titles */
			if (skip>MaxNamesTotal)
				header->size = MaxNamesTotal;
			if ((SongNames=(uint8*)FCEU_malloc(header->size))==0)
				return -2;
			if (FCEU_fread(SongNames,header->size,1,fp)!=1)
				goto error;
			skip-= MaxNamesTotal;
			break;
		case 'DATA':
			NSFSize = skip;
			if ((skip=NSFLoadData(fp))!=0)
				return skip;
			skip = 0;
			break;
		case 'NEND':
			return 1;
		case 'auth': /* cross our fingers about the encoding... */
			FCEU_fread(buf,MIN(skip,sizeof(buf)),1,fp);
			skip-= sizeof(buf);
			GetIndName(buf,NSFHeader.SongName,0);
			GetIndName(buf,NSFHeader.Artist,1);
			GetIndName(buf,NSFHeader.Copyright,2);
			GetIndName(buf,CaptionStr,3);
		}
		if (skip>0)
			FCEU_fseek(fp,skip,SEEK_CUR);
	}
	errs = "terminator tag (NEND) missing";
error:	FCEU_PrintError(" NSFe Error: %s.",
		errs);
	return -5;
}

int NSFLoad(FCEUFILE *fp)
{
	int		x;
	uint8		autoBS;

	*CaptionStr = 0;
	FCEU_fseek(fp,0,SEEK_SET);
	FCEU_fread(&NSFHeader,1,0x80,fp);
	if (memcmp(NSFHeader.ID,"NESM\x1A",5))
	{ /* check NSFe */
		if (memcmp(NSFHeader.ID,"NSFE",4))
			return 0;
		FCEU_fseek(fp,4,SEEK_SET);
		if ((x=NSFeLoad(fp))<=0)
			goto fail;
	} else
	{
		NSFSize = FCEU_fgetsize(fp)-0x80;
		if (FCEU_fseek(fp,0x80,SEEK_SET)<0)
			return 0;
	}
	/* does nothing if NSFe code already loaded */
	if ((x=NSFLoadData(fp))<=0) /* shouldn't fail */
		goto fail;
	NSFSizeLo = NSFSize&0xFF;
	NSFSizeHi = NSFSize>>8;

	NSFHeader.SongName[31] = NSFHeader.Artist[31] = NSFHeader.Copyright[31] = 0;

	--NSFMaxBank;

	NSFBankSW = x = autoBS = 0;
	do
		NSFBankSW|= NSFHeader.BankSwitch[x];
	while(++x<8);

	if (!NSFBankSW)
 		if ((NSFHeader.LoadAddressHigh&0x70)>=0x70)
			NSFBankSW = 0xFF; /* for Ice Climber and other F000 based NSFs */
		else
		{
			//autoBS = 1, BankCount = 0;
			//for(x=(NSFHeader.LoadAddressHigh&0x70)>>4; x<8; ++x)
			//	NSFHeader.BankSwitch[x] = BankCount, ++BankCount;
			//x = 0;
			//do
			//	NSFBankSW|= NSFHeader.BankSwitch[x];
			//while(++x<8);
		}

	FCEUGameInfo->type = GIT_NSF;
	FCEUGameInfo->input[0] = FCEUGameInfo->input[1] = SI_GAMEPAD;
	FCEUGameInfo->cspecial = SIS_NSF;

	if (!NSFHeader.VideoSystem)
		FCEUGameInfo->vidsys = GIV_NTSC;
	else if (NSFHeader.VideoSystem==1)
		FCEUGameInfo->vidsys = GIV_PAL;
	//otherwise user-defined

	GameInterface = NSFGI;

	FCEU_printf(
"NSF Loaded.  File information:\n"
//"\n"
" Name:          %s\n"
" Artist:        %s\n"
" Copyright:     %s\n",
		NSFHeader.SongName,NSFHeader.Artist,NSFHeader.Copyright);
	if (*CaptionStr)
		FCEU_printf(
" Rip Credit:    %s\n\n",
			CaptionStr);
	if (NSFHeader.SoundChip)
	{
		static char	*tab[6] = {
			"Konami VRC-6",
			"Konami VRC-7",
			"Nintendo FDS",
			"Nintendo MMC5",
			"Namco 106/163",
			"Sunsoft FME-07"
		};

		int		m;

		x = 0, m =-1;
		do
			if (NSFHeader.SoundChip&1<<x)
			{
				if (!++m)
					FCEU_printf(" Expansion hardware:  %s\n",
						tab[x]);
				else
					FCEU_printf("                    + %s\n",
						tab[x]);
				/* Prevent confusing weirdness if more than one bit is set. */
				// actually, a working combo may be used; 2010.03.18
				//r = 1<<x;
			}
		while(++x<6);
		//if (m>0)
		//{
		//	FCEU_printf("   Note: only the last one is used.\n");
		//	NSFHeader.SoundChip = r;
		//}
				
	}

	if (NSFBankSW)
		FCEU_printf(autoBS ?
" Bank-switched (auto)\n"
:" Bank-switched (%02X %02X %02X %02X %02X %02X %02X %02X).\n",
			NSFHeader.BankSwitch[0],
			NSFHeader.BankSwitch[1],
			NSFHeader.BankSwitch[2],
			NSFHeader.BankSwitch[3],
			NSFHeader.BankSwitch[4],
			NSFHeader.BankSwitch[5],
			NSFHeader.BankSwitch[6],
			NSFHeader.BankSwitch[7]);
	FCEU_printf(
" Load address:  $%04X\n"
" Init address:  $%04X\n"
" Play address:  $%04X\n"
" %4u KB / %u KB  %s\n",
		LoadAddr,
		NSFHeader.InitAddressLow|(NSFHeader.InitAddressHigh<<8),
		NSFHeader.PlayAddressLow|(NSFHeader.PlayAddressHigh<<8),
		NSFSize+1023>>10,NSFMaxBank*4+4,
		NSFHeader.VideoSystem==1 ? "PAL":NSFHeader.VideoSystem ? "NTSC/PAL":"NTSC");

	if (!NSFHeader.TotalSongs)
		NSFHeader.TotalSongs = 1; //fix it; 2010.03.18
	NSFLastSong = NSFHeader.TotalSongs-1;

	FCEU_printf(" Starting song:  %d / %d\n\n",
		NSFHeader.StartingSong,NSFHeader.TotalSongs);

	if (NSFHeader.SoundChip&1<<SCNFDS)
		ExWRAM = FCEU_malloc(32768+8192);
	else
		ExWRAM = FCEU_malloc(8192);

	if (!ExWRAM)
	{
		x =-2;
	fail:	NSFGI(GI_CLOSE); /* frees everything */
		return x;
	}
	return 1;
}

static DECLFR(NSF_VectorRead) //FFFA
{
	if (NSFNMIFlags&(NMIPendReload|NMISongReload)==NMIPendReload|NMISongReload
			|| NSFNMIFlags&(NMIWrTrigger|NMIReset))
	{
		if (A==0xFFFD)
			if (!fceuindbg) //was not present!; 2010.03.18
				NSFNMIFlags&=~NMIReset;
		if (A&1)
			return 0x3E;
		if (A==0xFFFC)
			return 0x44;
		return 0x00;
	} else
		return CartBRF8(A);
}

//make it public, let the emulated program do it (ain't I a stinker); 2010.03.18
//refer to NSF.h on the offsets
static DECLFR(NSF_HeadRead) //3F00
{
	if ((A&=0x7F)==0)
	{
		A = NSFNMIFlags;
		if (!fceuindbg)
			NSFNMIFlags&=~NMISongReload;
		return A;
	} else if (A<=7)
	{
		//if (A==1 && !(NSFHeader.SoundChip&1<<SCNFDS))
		//{
		//	BWrite[0x4017](0x4017,0xC0);
		//	BWrite[0x4017](0x4017,0x40);
		//}
		A+= 0x80; //into the "reserved" song current and reload
	}
	return ((const uint8*)&NSFHeader)[A];
}

static DECLFR(NSF_RegRead) //3F80
{
	if (!(A&0x7F))
		return PAL;
	return A>>8; //currently undefined
}

static DECLFW(NSF_Write) //3Fx0
{
	if ((A&=15)==0) //0 - set
		NSFNMIFlags&=~3, NSFNMIFlags|= V&3;
	else if (!--A) //1
		NSFNMIFlags|= NMISongReload;
	else if (!--A) //2
		NSFSpecial = V;
	else if (!--A) //3
		NSFNMIFlags&=~NMIPendReload;
	else if (!--A) //4
		NSFNMIFlags&=~NMIWrTrigger;
	else if (!--A) //5
		NSFNMIFlags|= NMIWrTrigger;
	else if (--A<=1) //6-7 - ExRAM side erase
		FCEU_memsetl(A*1024+ExWRAM,0,1024);
	else //RAM page erase
		FCEU_memsetl(A*256+(RAM-2*256),0,256);
}

static DECLFW(NSF_BankWrite) //5FF6/8-5FFF standard
{
	if (!NSFBankSW)
		return;
	A&= 0xF;
	BANKSET(A*4096,V);
}

static void RenderSongNames(void)
{
	if (SongNames)
		GetIndName(SongNames,CurrentSongName,NSFCurrentSong);
	else
		*CurrentSongName = 0;
	sprintf((char*)CaptionStr,"Song: <%d/%d>",
		NSFCurrentSong+1,NSFHeader.TotalSongs);
}

void NSF_init(void)
{
	int32		a, x;

	NSFNMIFlags = NMIReset;
	ResetCartMapping();

	if (NSFHeader.SoundChip&1<<SCNFDS)
	{
		SetupCartPRGMapping(0,ExWRAM,32768+8192,1);
		setprg32(0x6000,0);
		setprg8(0xE000,4);
		FCEU_memsetl(ExWRAM,0,32768+8192);
		SetCartRead(0x6000,0xFFFF);
		SetWriteHandler(0x6000,0xDFFF,CartBW);
	} else
	{
		SetupCartPRGMapping(0,NSFDATA,(NSFMaxBank+1)*4096,0);
		SetupCartPRGMapping(1,ExWRAM,8192,1);
		setprg8r(1,0x6000,0);
		FCEU_memsetl(ExWRAM,0,8192);
		SetCartRead(0x6000,0xFFFF);
		SetWriteHandler(0x6000,0x7FFF,CartBW);
	}

	a = 0x8000, x = 0;
	if (NSFBankSW)
		do
		{
			if (NSFHeader.SoundChip&1<<SCNFDS && x>=6) //also do 6xxx-7xxx
				BANKSET(x<<12,NSFHeader.BankSwitch[x]);
			BANKSET(a,NSFHeader.BankSwitch[x]);
		} while(++x,(a+=0x1000)<0x10000);
	else //new approach, should work
		do
			BANKSET(a,x);
		while(x+=(a>=(LoadAddr&0xF000)),(a+=0x1000)<0x10000); //missing (); 2010.05.10

	SetReadHandler(0x2000,0x3DFF,0);
	SetReadHandler(0x3E00,0x3EFF,NSF_ROMRead);
	SetReadHandler(0x3F00,0x3F7F,NSF_HeadRead);
	SetReadHandler(0x3F80,0x3FFF,NSF_RegRead);
	SetReadHandler(0xFFFA,0xFFFD,NSF_VectorRead);
	SetWriteHandler(0x2000,0x3DFF,0);
	SetWriteHandler(0x3E00,0x3FFF,NSF_Write);
	SetWriteHandler(0x5FF6,0x5FF7,NSFHeader.SoundChip&1<<SCNFDS ? NSF_BankWrite:0); //was ? 0:..
	SetWriteHandler(0x5FF8,0x5FFF,NSF_BankWrite);

	if (NSFHeader.SoundChip&1<<SCVRC6)
		NSFVRC6_Init();
	if (NSFHeader.SoundChip&1<<SCVRC7)
		NSFVRC7_Init();
	if (NSFHeader.SoundChip&1<<SCNFDS)
		FDSSoundReset();
	if (NSFHeader.SoundChip&1<<SCMMC5)
		NSFMMC5_Init();
	if (NSFHeader.SoundChip&1<<SCN106)
		NSFN106_Init();
	if (NSFHeader.SoundChip&1<<SCSFME)
		NSFAY_Init();
	NSFCurrentSong = NSFHeader.StartingSong-1;
	NSFNMIFlags|= NMISongReload;
	NSFNMIFlags&=~(NMIPendReload|NMIWrTrigger);

	RenderSongNames();
}

static void TestJoy(void)
{
	static uint8	last = 0;
	uint8		current = FCEU_GetJoyJoy();

	if ((current&JOY_RIGHT) && !(last&JOY_RIGHT))
	{
		if (NSFCurrentSong<NSFLastSong)
		{
			++NSFCurrentSong;
			NSFNMIFlags|= NMISongReload;
		}
	} else if ((current&JOY_LEFT) && !(last&JOY_LEFT))
	{
		if (NSFCurrentSong>0)
		{
			--NSFCurrentSong;
			NSFNMIFlags|= NMISongReload;
		}
	} else if ((current&JOY_UP) && !(last&JOY_UP))
	{
		if (NSFCurrentSong+10>NSFLastSong)
			NSFCurrentSong = NSFLastSong;
		else
			NSFCurrentSong+= 10;
		NSFNMIFlags|= NMISongReload;
	} else if ((current&JOY_DOWN) && !(last&JOY_DOWN))
	{
		if (NSFCurrentSong<10)
			NSFCurrentSong = 0;
		else
			NSFCurrentSong-= 10;
		NSFNMIFlags|= NMISongReload;
	} else if ((current&JOY_START) && !(last&JOY_START))
		NSFNMIFlags|= NMISongReload;
	else if ((current&JOY_A) && !(last&JOY_A))
		NSFSpecial = (NSFSpecial+1)%3;
	else
		goto skip;
	RenderSongNames();
skip:
	last = current;
}

#define ITERATION_FRACTION	(0x80000000U/256)
#define FULL_CIRCLE		0

void DrawNSF(uint8 *XBuf)
{
	int16		*Bufpl;
	int32		mul;
	int		special;
	int		l, x = 0, i = 0;

	if (!vismode)
		return;
	if (!XBuf)
		goto input;

	/* use void color (if selected); 2009.12.29 */
	FCEU_memsetl(XBuf,
		PPUvoidcolor==0xFF ? 0:(l=PPUvoidcolor,l|=l<<8,l|=l<<16),
		256*240);

	l = GetSoundBuffer(&Bufpl);

	if (!FSettings.SoundVolume)
		mul = 0;
	else
		mul = (50*240)/FSettings.SoundVolume;

	if ((special=NSFSpecial&3)==0)
	{
		do
		{
			uint32		y = 142+((Bufpl[i>>8]*mul)>>14);

			if (y<240)
				XBuf[x+y*256] = 3;
		} while(i+=l,++x<256);
	} else if (special==1)
	{
		ufract		t = 0;

		do
		{ /* all of the rest of the trig functions used can be replaced */
			int32		r = (Bufpl[i>>8]*mul)>>14;
			uint32		xp = 128+cosmul(r,t);
			uint32		yp = 120+sinmul(r,t);

			xp&= 255;
			yp%= 240;
			XBuf[xp+yp*256] = 3;
		} while(i+=l,++x,(t+=ITERATION_FRACTION*2)!=FULL_CIRCLE);
	} else if (special==2)
	{
		static ufract	theta = 0;

		x = 0;
		do
		{
			int32		xc = 128-x;
			int32		yc = 0-(int32)((Bufpl[i>>8]*mul)>>14);
			ufract		t = theta+0x80000000U+ifractatan2(yc,xc);
			int32		r = ihypot2(xc,yc);
			uint32		m = 128+cosmul(r,t);
			uint32		n = 120+sinmul(r,t);

			if (m<256 && n<240)
				XBuf[m+n*256] = 3;
		} while(i+=l,++x<128);

		do
		{
			int32		xc = x-128;
			int32		yc = (Bufpl[i>>8]*mul)>>14;
			ufract		t = theta+ifractatan2(yc,xc);
			int32		r = ihypot2(xc,yc);
			uint32		m = 128+cosmul(r,t);
			uint32		n = 120+sinmul(r,t);

			if (m<256 && n<240)
				XBuf[m+n*256] = 3;
		} while(i+=l,++x<256);
		theta += ITERATION_FRACTION;
	}

	DrawTextTrans(XBuf+10*256+4+(31-strlen((char*)NSFHeader.SongName)<<2),
		256, NSFHeader.SongName, 6);
/*26*/	DrawTextTrans(XBuf+25*256+4+(31-strlen((char*)NSFHeader.Artist)<<2),
		256,NSFHeader.Artist, 6);
/*42*/	DrawTextTrans(XBuf+40*256+4+(31-strlen((char*)NSFHeader.Copyright)<<2),
		256,NSFHeader.Copyright, 6);
/*70*/	DrawTextTrans(XBuf+55*256+4+(31-strlen((char*)CaptionStr)<<2),
		256,CaptionStr, 6);
	if (*CurrentSongName)
		DrawTextTrans(XBuf+70*256+4+(31-strlen((char*)CurrentSongName)<<2),
			256,CurrentSongName, 6);
input:	TestJoy();
}

void DoNSFFrame(void)
{
	if ((NSFNMIFlags&(NMIPendReload|NMISongReload)==NMIPendReload|NMISongReload)
			|| NSFNMIFlags&NMIWrTrigger)
		XFrame.IRQlow|= FCEU_IQNMI; //TriggerNMI
}

void FCEUI_NSFSetVis(int mode)
{
	vismode = !!mode;
}

int FCEUI_NSFChange(int amount)
{
	int	current = NSFCurrentSong;

	if ((current+=amount)<0)
		current = 0;
	else if (current>NSFLastSong)
		current = NSFLastSong;
	NSFNMIFlags|= NMISongReload;

	return NSFCurrentSong=current;
}

/* Returns total songs */
int FCEUI_NSFGetInfo(uint8 *name,
	uint8 *artist,
	uint8 *copyright,
	int maxlen)
{
	strncpy((char*)name,(char*)NSFHeader.SongName,maxlen);
	strncpy((char*)artist,(char*)NSFHeader.Artist,maxlen);
	strncpy((char*)copyright,(char*)NSFHeader.Copyright,maxlen);

	return NSFHeader.TotalSongs;
}
