/***********************************************************************************

  emu2413.c -- YM2413 emulator written by Mitsutaka Okazaki 2001

  2001 01-08 : Version 0.10 -- 1st version.
  2001 01-15 : Version 0.20 -- semi-public version.
  2001 01-16 : Version 0.30 -- 1st public version.
  2001 01-17 : Version 0.31 -- Fixed bassdrum problem.
             : Version 0.32 -- LPF implemented.
  2001 01-18 : Version 0.33 -- Fixed the drum problem, refine the mix-down method.
                            -- Fixed the LFO bug.
  2001 01-24 : Version 0.35 -- Fixed the drum problem,
                               support undocumented EG behavior.
  2001 02-02 : Version 0.38 -- Improved the performance.
                               Fixed the hi-hat and cymbal model.
                               Fixed the default percussive datas.
                               Noise reduction.
                               Fixed the feedback problem.
  2001 03-03 : Version 0.39 -- Fixed some drum bugs.
                               Improved the performance.
  2001 03-04 : Version 0.40 -- Improved the feedback.
                               Change the default table size.
                               Clock and Rate can be changed during play.
  2001 06-24 : Version 0.50 -- Improved the hi-hat and the cymbal tone.
                               Added VRC7 patch (OPLL_reset_patch is changed).
                               Fixed OPLL_reset() bug.
                               Added OPLL_setMask, OPLL_getMask and OPLL_toggleMask.
                               Added OPLL_writeIO.
  2001 09-28 : Version 0.51 -- Removed the noise table.
  2002 01-28 : Version 0.52 -- Added Stereo mode.
  2002 02-07 : Version 0.53 -- Fixed some drum bugs.
  2002 02-20 : Version 0.54 -- Added the best quality mode.
  2002 03-02 : Version 0.55 -- Removed OPLL_init & OPLL_close.
  2002 05-30 : Version 0.60 -- Fixed HH&CYM generator and all voice datas.

  2004 01-24 : Modified by xodnizel to remove code not needed for the VRC7, among other things.

  2013 05-14 : Modified to remove obsolete end-shifts and FP-laden redundant table building.
  		The moocow funcs appropriately renamed fill2413.

  References:
    fmopl.c        -- 1999,2000 written by Tatsuyuki Satoh (MAME development).
    fmopl.c(fixed) -- (C) 2002 Jarek Burczynski.
    s_opl.c        -- 2001 written by Mamiya (NEZplug development).
    fmgen.cpp      -- 1999,2000 written by cisc.
    fmpac.ill      -- 2000 created by NARUTO.
    MSX-Datapack
    YMU757 data sheet
    YM2143 data sheet

**************************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "emu2413.h"

static const unsigned char default_inst[15][8] = {
 #include "vrc7tone.h"
};

/* Size of Sintable ( 8 -- 18 can be used. 9 recommended.)*/
#define PG_BITS 9
#define PG_WIDTH (1<<PG_BITS)

/* Phase increment counter */
#define DP_BITS 18
#define DP_WIDTH (1<<DP_BITS)
#define DP_BASE_BITS (DP_BITS - PG_BITS)

/* Dynamic range (Accuracy of sin table) */
#define DB_BITS 8
#define DB_STEP (48.0/(1<<DB_BITS))
#define DB_MUTE (1<<DB_BITS)

/* Dynamic range of envelope */
#define EG_STEP 0.375
#define EG_BITS 7
#define EG_WIDTH (1<<EG_BITS)
#define EG_MUTE EG_WIDTH

/* Dynamic range of total level */
#define TL_STEP 0.75
#define TL_BITS 6
#define TL_MUTE (1<<TL_BITS)

/* Dynamic range of sustine level */
#define SL_STEP 3.0
#define SL_BITS 4
#define SL_MUTE (1<<SL_BITS)

#define EG2DB(d) ((d)*(int32)(EG_STEP/DB_STEP))
#define TL2EG(d) ((d)*(int32)(TL_STEP/EG_STEP))
#define SL2EG(d) ((d)*(int32)(SL_STEP/EG_STEP))

#define DB_POS(x) (uint32)((x)/DB_STEP)
#define DB_NEG(x) (uint32)(DB_MUTE+DB_MUTE+(x)/DB_STEP)

/* Bits for liner value */
#define DB2LIN_AMP_BITS 11
#define SLOT_AMP_BITS (DB2LIN_AMP_BITS)

/* Bits for envelope phase incremental counter */
#define EG_DP_BITS 22
#define EG_DP_WIDTH (1<<EG_DP_BITS)

/* Bits for Pitch and Amp modulator */
#define PM_PG_BITS 8
#define PM_PG_WIDTH (1<<PM_PG_BITS)
#define PM_DP_BITS 16
#define PM_DP_WIDTH (1<<PM_DP_BITS)
#define AM_PG_BITS 8
#define AM_PG_WIDTH (1<<AM_PG_BITS)
#define AM_DP_BITS 16
#define AM_DP_WIDTH (1<<AM_DP_BITS)

/* PM table is calcurated by PM_AMP * pow(2,PM_DEPTH*sin(x)/1200) */
#define PM_AMP_BITS 8
#define PM_AMP (1<<PM_AMP_BITS)

/* PM speed(Hz) and depth(cent) */
#define PM_SPEED 6.4
#define PM_DEPTH 13.75

/* AM speed(Hz) and depth(dB) */
#define AM_SPEED 3.7
//#define AM_DEPTH 4.8
#define AM_DEPTH 2.4

/* Cut the lower b bit(s) off. */
#define HIGHBITS(c,b) ((c)>>(b))

/* Leave the lower b bit(s). */
#define LOWBITS(c,b) ((c)&((1<<(b))-1))

/* Expand x which is s bits to d bits. */
#define EXPAND_BITS(x,s,d) ((x)<<((d)-(s)))

/* Expand x which is s bits to d bits and fill expanded bits '1' */
#define EXPAND_BITS_X(x,s,d) (((x)<<((d)-(s)))|((1<<((d)-(s)))-1))

/* Adjust envelope speed which depends on sampling rate. */
#define rate_adjust(x) (rate==49716?x:(uint32)((double)(x)*clk/72/rate + 0.5))        /* added 0.5 to round the value*/

#define MOD(o,x) (&(o)->slot[(x)<<1])
#define CAR(o,x) (&(o)->slot[((x)<<1)|1])

#define BIT(s,b) (((s)>>(b))&1)

/* Input clock */
static uint32 clk = 844451141;
/* Sampling rate */
static uint32 rate = 3354932;

/* WaveTable for each envelope amp */
static uint16 fullsintable[PG_WIDTH]
#if PG_WIDTH==512
= {
	255,203,171,152,139,129,120,113,107,102,97,92,88,85,81,78,
	75,72,70,67,65,63,61,59,57,55,53,52,50,48,47,45,
	44,43,41,40,39,38,37,35,34,33,32,31,30,29,28,28,
	27,26,25,24,23,23,22,21,21,20,19,19,18,17,17,16,
	16,15,14,14,13,13,12,12,11,11,11,10,10,9,9,8,
	8,8,7,7,7,6,6,6,5,5,5,4,4,4,4,3,
	3,3,3,2,2,2,2,2,2,1,1,1,1,1,1,1
}
#endif
;
static uint16 halfsintable[PG_WIDTH];

static uint16 *waveform[2] = { fullsintable, halfsintable };

/* LFO Table */
#if PM_PG_WIDTH==256
static int16 pmtable[PM_PG_WIDTH] = {
	256,256,256,256,256,256,256,256,256,256,256,256,256,256,256,256,
	256,256,256,256,256,257,257,257,257,257,257,257,257,257,257,257,
	257,257,257,257,257,257,257,257,257,257,257,257,257,257,257,257,
	257,257,257,257,257,257,257,257,258,258,258,258,258,258,258,258,
	258,258,258,258,258,258,258,258,258,257,257,257,257,257,257,257,
	257,257,257,257,257,257,257,257,257,257,257,257,257,257,257,257,
	257,257,257,257,257,257,257,257,257,257,257,257,256,256,256,256,
	256,256,256,256,256,256,256,256,256,256,256,256,256,256,256,256,
	256,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,254,254,254,254,254,254,254,254,254,254,254,
	254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,
	254,254,254,254,254,254,254,254,254,254,253,253,253,253,253,253,
	253,253,253,253,253,253,253,254,254,254,254,254,254,254,254,254,
	254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,254,
	254,254,254,254,254,254,254,254,254,254,254,254,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255
};
#else
static int32 pmtable[PM_PG_WIDTH];
#endif

#if AM_PG_WIDTH==256
static int8	amtable[AM_PG_WIDTH] = {
	6,6,6,6,7,7,7,7,7,7,7,8,8,8,8,8,
	8,8,9,9,9,9,9,9,9,10,10,10,10,10,10,10,
	10,11,11,11,11,11,11,11,11,11,11,11,12,12,12,12,
	12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,
	12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,
	12,12,12,12,12,11,11,11,11,11,11,11,11,11,11,11,
	10,10,10,10,10,10,10,10,9,9,9,9,9,9,9,8,
	8,8,8,8,8,8,7,7,7,7,7,7,7,6,6,6,
	6,6,6,5,5,5,5,5,5,4,4,4,4,4,4,4,
	3,3,3,3,3,3,3,2,2,2,2,2,2,2,2,1,
	1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,
	1,1,2,2,2,2,2,2,2,2,3,3,3,3,3,3,
	3,4,4,4,4,4,4,4,5,5,5,5,5,5,6,6
};
#else
static int32 amtable[AM_PG_WIDTH];
#endif

/* Phase delta for LFO */
static uint32 pm_dphase;
static uint32 am_dphase;

/* dB to Liner table */
#if ((DB_MUTE + DB_MUTE) * 2)==1024
static int16 DB2LIN_TABLE[1024] = {
	2047,2003,1960,1918,1877,1837,1798,1759,1722,1685,1649,1614,1579,1546,1513,1480,
	1449,1418,1387,1358,1329,1300,1273,1245,1219,1193,1167,1142,1118,1094,1071,1048,
	1025,1004,982,961,941,920,901,882,863,844,826,809,791,774,758,742,
	726,710,695,680,666,651,638,624,611,598,585,572,560,548,536,525,
	514,503,492,481,471,461,451,442,432,423,414,405,396,388,380,371,
	364,356,348,341,333,326,319,312,306,299,293,287,280,274,269,263,
	257,252,246,241,236,231,226,221,216,212,207,203,198,194,190,186,
	182,178,174,170,167,163,160,156,153,150,147,143,140,137,134,131,
	129,126,123,121,118,115,113,111,108,106,104,101,99,97,95,93,
	91,89,87,85,83,82,80,78,76,75,73,72,70,69,67,66,
	64,63,61,60,59,58,56,55,54,53,52,51,49,48,47,46,
	45,44,43,42,42,41,40,39,38,37,36,36,35,34,33,33,
	32,31,31,30,29,29,28,27,27,26,26,25,25,24,23,23,
	22,22,21,21,21,20,20,19,19,18,18,18,17,17,16,16,
	16,15,15,15,14,14,14,13,13,13,13,12,12,12,12,11,
	11,11,11,10,10,10,10,9,9,9,9,9,8,8,8,8,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	-2047,-2003,-1960,-1918,-1877,-1837,-1798,-1759,-1722,-1685,-1649,-1614,-1579,-1546,-1513,-1480,
	-1449,-1418,-1387,-1358,-1329,-1300,-1273,-1245,-1219,-1193,-1167,-1142,-1118,-1094,-1071,-1048,
	-1025,-1004,-982,-961,-941,-920,-901,-882,-863,-844,-826,-809,-791,-774,-758,-742,
	-726,-710,-695,-680,-666,-651,-638,-624,-611,-598,-585,-572,-560,-548,-536,-525,
	-514,-503,-492,-481,-471,-461,-451,-442,-432,-423,-414,-405,-396,-388,-380,-371,
	-364,-356,-348,-341,-333,-326,-319,-312,-306,-299,-293,-287,-280,-274,-269,-263,
	-257,-252,-246,-241,-236,-231,-226,-221,-216,-212,-207,-203,-198,-194,-190,-186,
	-182,-178,-174,-170,-167,-163,-160,-156,-153,-150,-147,-143,-140,-137,-134,-131,
	-129,-126,-123,-121,-118,-115,-113,-111,-108,-106,-104,-101,-99,-97,-95,-93,
	-91,-89,-87,-85,-83,-82,-80,-78,-76,-75,-73,-72,-70,-69,-67,-66,
	-64,-63,-61,-60,-59,-58,-56,-55,-54,-53,-52,-51,-49,-48,-47,-46,
	-45,-44,-43,-42,-42,-41,-40,-39,-38,-37,-36,-36,-35,-34,-33,-33,
	-32,-31,-31,-30,-29,-29,-28,-27,-27,-26,-26,-25,-25,-24,-23,-23,
	-22,-22,-21,-21,-21,-20,-20,-19,-19,-18,-18,-18,-17,-17,-16,-16,
	-16,-15,-15,-15,-14,-14,-14,-13,-13,-13,-13,-12,-12,-12,-12,-11,
	-11,-11,-11,-10,-10,-10,-10,-9,-9,-9,-9,-9,-8,-8,-8,-8
};
#else
static int16 DB2LIN_TABLE[(DB_MUTE + DB_MUTE) * 2];
#endif

/* Liner to Log curve conversion table (for Attack rate). */
#if EG_WIDTH==128
static uint8	AR_ADJUST_TABLE[128] = {
	128,127,108,98,90,84,79,75,72,69,66,63,61,59,57,55,
	53,52,50,49,47,46,45,44,43,42,41,40,39,38,37,36,
	35,34,33,33,32,31,31,30,29,29,28,27,27,26,25,25,
	24,24,23,23,22,22,21,21,20,20,19,19,18,18,18,17,
	17,16,16,16,15,15,14,14,14,13,13,13,12,12,12,11,
	11,11,10,10,10,9,9,9,8,8,8,8,7,7,7,6,
	6,6,6,5,5,5,4,4,4,4,3,3,3,3,2,2,
	2,2,2,1,1,1,1
};
#else
static uint16 AR_ADJUST_TABLE[EG_WIDTH];
#endif

/* Definition of envelope mode */
enum
{ SETTLE, ATTACK, DECAY, SUSHOLD, SUSTINE, RELEASE, FINISH };

/* Phase incr table for Attack */
static uint32 dphaseARTable[16][16];
/* Phase incr table for Decay and Release */
static uint32 dphaseDRTable[16][16];

/* KSL + TL Table */
static uint32 tllTable[16][8][1 << TL_BITS][4];
static int32 rksTable[2][8][2];

/* Phase incr table for PG */
static uint32 dphaseTable[512][8][16];

/***************************************************

                  Create tables

****************************************************/
INLINE static int32 Min (int32 i,
	int32 j)
{
  if(i < j)
    return i;
  else
    return j;
}

/* Table for AR to LogCurve. */
#if EG_WIDTH==128
 #define makeAdjustTable()
#else
static void makeAdjustTable (void)
{
  int32 i;

  AR_ADJUST_TABLE[0] = EG_MUTE;
  for (i = 1; i < EG_WIDTH; i++) /* was <128 */
    AR_ADJUST_TABLE[i] = (uint16) ((double) (1 << EG_BITS) - 1 - (1 << EG_BITS) * log (i) / log (128));
}
#endif

/* Table for dB(0 -- (1<<DB_BITS)-1) to Liner(0 -- DB2LIN_AMP_WIDTH) */
#if ((DB_MUTE+DB_MUTE)*2)==1024
 #define makeDB2LinTable()
#else
static void makeDB2LinTable (void)
{
  int32 i;

  for (i = 0; i < DB_MUTE + DB_MUTE; i++)
  {
    DB2LIN_TABLE[i] = (int16) ((double) ((1 << DB2LIN_AMP_BITS) - 1) * pow (10, -(double) i * DB_STEP / 20));
    if(i >= DB_MUTE) DB2LIN_TABLE[i] = 0;
    DB2LIN_TABLE[i + DB_MUTE + DB_MUTE] = (int16) (-DB2LIN_TABLE[i]);
  }
}
#endif

/* Liner(+0.0 - +1.0) to dB((1<<DB_BITS) - 1 -- 0) */
static int32 lin2db (double d)
{
  if(d == 0)
    return (DB_MUTE - 1);
  else
    return Min (-(int32) (20.0 * log10 (d) / DB_STEP), DB_MUTE-1);  /* 0 -- 127 */
}


/* Sin Table */
static void makeSinTable (void)
{
  int32 i;
#if PG_WIDTH!=512
  for (i = 0; i < PG_WIDTH / 4; i++)
  {
    fullsintable[i] = (uint32) lin2db (sin (2.0 * PI * i / PG_WIDTH) );
  }
#endif
  for (i = 0; i < PG_WIDTH / 4; i++)
  {
    fullsintable[PG_WIDTH / 2 - 1 - i] = fullsintable[i];
  }

  for (i = 0; i < PG_WIDTH / 2; i++)
  {
    fullsintable[PG_WIDTH / 2 + i] = (uint32) (DB_MUTE + DB_MUTE + fullsintable[i]);
  }

  for (i = 0; i < PG_WIDTH / 2; i++)
    halfsintable[i] = fullsintable[i];
  for (i = PG_WIDTH / 2; i < PG_WIDTH; i++)
    halfsintable[i] = fullsintable[0];
}

/* Table for Pitch Modulator */
#if PM_PG_WIDTH==256
 #define makePmTable()
#else
static void makePmTable (void)
{
  int32 i;

  for (i = 0; i < PM_PG_WIDTH; i++)
    pmtable[i] = (int32) ((double) PM_AMP * pow (2, (double) PM_DEPTH * sin (2.0 * PI * i / PM_PG_WIDTH) / 1200));
}
#endif

/* Table for Amp Modulator */
#if AM_PG_WIDTH==256
 #define makeAmTable()
#else
static void makeAmTable (void)
{
  int32 i;

  for (i = 0; i < AM_PG_WIDTH; i++)
    amtable[i] = (int32) ((double) AM_DEPTH / 2 / DB_STEP * (1.0 + sin (2.0 * PI * i / PM_PG_WIDTH)));
}
#endif

/* Phase increment counter table */
static void makeDphaseTable (void)
{
  uint32 fnum, block, ML;
  uint32 mltable[16] =
    { 1, 1 * 2, 2 * 2, 3 * 2, 4 * 2, 5 * 2, 6 * 2, 7 * 2, 8 * 2, 9 * 2, 10 * 2, 10 * 2, 12 * 2, 12 * 2, 15 * 2, 15 * 2 };

  for (fnum = 0; fnum < 512; fnum++)
    for (block = 0; block < 8; block++)
      for (ML = 0; ML < 16; ML++)
        dphaseTable[fnum][block][ML] = rate_adjust (((fnum * mltable[ML]) << block) >> (20 - DP_BITS));
}

static void makeTllTable (void)
{
#define dB2(x) ((x)*2)

  static int8 kltable[16] = { /* was double - truncated in routine */
    dB2 (0.000), dB2 (9.000), dB2 (12.000), dB2 (13.875), dB2 (15.000), dB2 (16.125), dB2 (16.875), dB2 (17.625),
    dB2 (18.000), dB2 (18.750), dB2 (19.125), dB2 (19.500), dB2 (19.875), dB2 (20.250), dB2 (20.625), dB2 (21.000)
  };

  int32 tmp;
  int32 fnum, block, TL, KL;

  for (fnum = 0; fnum < 16; fnum++)
    for (block = 0; block < 8; block++)
      for (TL = 0; TL < (1<<TL_BITS); TL++) /* was <64 */
        for (KL = 0; KL < 4; KL++)
        {
          if(KL == 0)
          {
            tllTable[fnum][block][TL][KL] = TL2EG (TL);
          }
          else
          {
            tmp = (int32) (kltable[fnum] - dB2 (3.000) * (7 - block));
            if(tmp <= 0)
              tllTable[fnum][block][TL][KL] = TL2EG (TL);
            else
              tllTable[fnum][block][TL][KL] = (uint32) ((tmp >> (3 - KL)) / EG_STEP) + TL2EG (TL);
          }
        }
}

#ifdef USE_SPEC_ENV_SPEED

#define RTIME(v) (1<<EG_DP_BITS)/((v)*3579545/72000)

static uint32 attacktable[16][4] = {
  {0, 0, 0, 0},
  {RTIME(1730.15), RTIME(1400.60), RTIME(1153.43), RTIME(988.66)},
  {RTIME(865.08), RTIME(700.30), RTIME(576.72), RTIME(494.33)},
  {RTIME(432.54), RTIME(350.15), RTIME(288.36), RTIME(247.16)},
  {RTIME(216.27), RTIME(175.07), RTIME(144.18), RTIME(123.58)},
  {RTIME(108.13), RTIME(87.54), RTIME(72.09), RTIME(61.79)},
  {RTIME(54.07), RTIME(43.77), RTIME(36.04), RTIME(30.90)},
  {RTIME(27.03), RTIME(21.88), RTIME(18.02), RTIME(15.45)},
  {RTIME(13.52), RTIME(10.94), RTIME(9.01), RTIME(7.72)},
  {RTIME(6.76), RTIME(5.47), RTIME(4.51), RTIME(3.86)},
  {RTIME(3.38), RTIME(2.74), RTIME(2.25), RTIME(1.93)},
  {RTIME(1.69), RTIME(1.37), RTIME(1.13), RTIME(0.97)},
  {RTIME(0.84), RTIME(0.70), RTIME(0.60), RTIME(0.54)},
  {RTIME(0.50), RTIME(0.42), RTIME(0.34), RTIME(0.30)},
  {RTIME(0.28), RTIME(0.22), RTIME(0.18), RTIME(0.14)},
  {EG_DP_WIDTH, EG_DP_WIDTH, EG_DP_WIDTH, EG_DP_WIDTH}
};

static uint32 decaytable[16][4] = {
  {0, 0, 0, 0},
  {RTIME(20926.60), RTIME(16807.20), RTIME(14006.00), RTIME(12028.60)},
  {RTIME(10463.30), RTIME(8403.58), RTIME(7002.98), RTIME(6014.32)},
  {RTIME(5231.64), RTIME(4201.79), RTIME(3501.49), RTIME(3007.16)},
  {RTIME(2615.82), RTIME(2100.89), RTIME(1750.75), RTIME(1503.58)},
  {RTIME(1307.91), RTIME(1050.45), RTIME(875.37), RTIME(751.79)},
  {RTIME(653.95), RTIME(525.22), RTIME(437.69), RTIME(375.90)},
  {RTIME(326.98), RTIME(262.61), RTIME(218.84), RTIME(187.95)},
  {RTIME(163.49), RTIME(131.31), RTIME(109.42), RTIME(93.97)},
  {RTIME(81.74), RTIME(65.65), RTIME(54.71), RTIME(46.99)},
  {RTIME(40.87), RTIME(32.83), RTIME(27.36), RTIME(23.49)},
  {RTIME(20.44), RTIME(16.41), RTIME(13.68), RTIME(11.75)},
  {RTIME(10.22), RTIME(8.21), RTIME(6.84), RTIME(5.87)},
  {RTIME(5.11), RTIME(4.10), RTIME(3.42), RTIME(2.94)},
  {RTIME(2.55), RTIME(2.05), RTIME(1.71), RTIME(1.47)},
  {RTIME(1.27), RTIME(1.27), RTIME(1.27), RTIME(1.27)}
};
#endif

/* Rate Table for Attack */
static void
makeDphaseARTable (void)
{
  int32 AR, Rks, RM, RL;

  for (AR = 0; AR < 16; AR++)
    for (Rks = 0; Rks < 16; Rks++)
    {
      RM = AR + (Rks >> 2);
      RL = Rks & 3;
      if(RM > 15)
        RM = 15;
      switch (AR)
      {
      case 0:
        dphaseARTable[AR][Rks] = 0;
        break;
      case 15:
        dphaseARTable[AR][Rks] = 0;/*EG_DP_WIDTH;*/
        break;
      default:
#ifdef USE_SPEC_ENV_SPEED
        dphaseARTable[AR][Rks] = rate_adjust (attacktable[RM][RL]);
#else
        dphaseARTable[AR][Rks] = rate_adjust ((3 * (RL + 4) << (RM + 1)));
#endif
        break;
      }
    }
}

/* Rate Table for Decay and Release */
static void
makeDphaseDRTable (void)
{
  int32 DR, Rks, RM, RL;

  for (DR = 0; DR < 16; DR++)
    for (Rks = 0; Rks < 16; Rks++)
    {
      RM = DR + (Rks >> 2);
      RL = Rks & 3;
      if(RM > 15)
        RM = 15;
      switch (DR)
      {
      case 0:
        dphaseDRTable[DR][Rks] = 0;
        break;
      default:
#ifdef USE_SPEC_ENV_SPEED
        dphaseDRTable[DR][Rks] = rate_adjust (decaytable[RM][RL]);
#else
        dphaseDRTable[DR][Rks] = rate_adjust ((RL + 4) << (RM - 1));
#endif
        break;
      }
    }
}

static void
makeRksTable (void)
{

  int32 fnum8, block, KR;

  for (fnum8 = 0; fnum8 < 2; fnum8++)
    for (block = 0; block < 8; block++)
      for (KR = 0; KR < 2; KR++)
      {
        if(KR != 0)
          rksTable[fnum8][block][KR] = (block << 1) + fnum8;
        else
          rksTable[fnum8][block][KR] = block >> 1;
      }
}

/************************************************************

                      Calc Parameters

************************************************************/

INLINE static uint32
calc_eg_dphase (OPLL_SLOT * slot)
{

  switch (slot->eg_mode)
  {
  case ATTACK:
    return dphaseARTable[slot->patch.AR][slot->rks];

  case DECAY:
    return dphaseDRTable[slot->patch.DR][slot->rks];

  case SUSHOLD:
    return 0;

  case SUSTINE:
    return dphaseDRTable[slot->patch.RR][slot->rks];

  case RELEASE:
    if(slot->sustine)
      return dphaseDRTable[5][slot->rks];
    else if(slot->patch.EG)
      return dphaseDRTable[slot->patch.RR][slot->rks];
    else
      return dphaseDRTable[7][slot->rks];

  case FINISH:
    return 0;

  default:
    return 0;
  }
}

/*************************************************************

                    OPLL internal interfaces

*************************************************************/

#define UPDATE_PG(S)  (S)->dphase = dphaseTable[(S)->fnum][(S)->block][(S)->patch.ML]
#define UPDATE_TLL(S)\
(((S)->type==0)?\
((S)->tll = tllTable[((S)->fnum)>>5][(S)->block][(S)->patch.TL][(S)->patch.KL]):\
((S)->tll = tllTable[((S)->fnum)>>5][(S)->block][(S)->volume][(S)->patch.KL]))
#define UPDATE_RKS(S) (S)->rks = rksTable[((S)->fnum)>>8][(S)->block][(S)->patch.KR]
#define UPDATE_WF(S)  (S)->sintbl = waveform[(S)->patch.WF]
#define UPDATE_EG(S)  (S)->eg_dphase = calc_eg_dphase(S)
#define UPDATE_ALL(S)\
  UPDATE_PG(S);\
  UPDATE_TLL(S);\
  UPDATE_RKS(S);\
  UPDATE_WF(S); \
  UPDATE_EG(S)                  /* EG should be updated last. */


/* Slot key on  */
INLINE static void
slotOn (OPLL_SLOT * slot)
{
  slot->eg_mode = ATTACK;
  slot->eg_phase = 0;
  slot->phase = 0;
}

/* Slot key on without reseting the phase */
INLINE static void
slotOn2 (OPLL_SLOT * slot)
{
  slot->eg_mode = ATTACK;
  slot->eg_phase = 0;
}

/* Slot key off */
INLINE static void
slotOff (OPLL_SLOT * slot)
{
  if(slot->eg_mode == ATTACK)
    slot->eg_phase = EXPAND_BITS (AR_ADJUST_TABLE[HIGHBITS (slot->eg_phase, EG_DP_BITS - EG_BITS)], EG_BITS, EG_DP_BITS);
  slot->eg_mode = RELEASE;
}

/* Channel key on */
INLINE static void
keyOn (OPLL * opll, int32 i)
{
  if(!opll->slot_on_flag[i * 2])
    slotOn (MOD(opll,i));
  if(!opll->slot_on_flag[i * 2 + 1])
    slotOn (CAR(opll,i));
  opll->key_status[i] = 1;
}

/* Channel key off */
INLINE static void
keyOff (OPLL * opll, int32 i)
{
  if(opll->slot_on_flag[i * 2 + 1])
    slotOff (CAR(opll,i));
  opll->key_status[i] = 0;
}

/* Set sustine parameter */
INLINE static void
setSustine (OPLL * opll, int32 c, int32 sustine)
{
  CAR(opll,c)->sustine = sustine;
  if(MOD(opll,c)->type)
    MOD(opll,c)->sustine = sustine;
}

/* Volume : 6bit ( Volume register << 2 ) */
INLINE static void
setVolume (OPLL * opll, int32 c, int32 volume)
{
  CAR(opll,c)->volume = volume;
}

INLINE static void
setSlotVolume (OPLL_SLOT * slot, int32 volume)
{
  slot->volume = volume;
}

/* Set F-Number ( fnum : 9bit ) */
INLINE static void
setFnumber (OPLL * opll, int32 c, int32 fnum)
{
  CAR(opll,c)->fnum = fnum;
  MOD(opll,c)->fnum = fnum;
}

/* Set Block data (block : 3bit ) */
INLINE static void
setBlock (OPLL * opll, int32 c, int32 block)
{
  CAR(opll,c)->block = block;
  MOD(opll,c)->block = block;
}

INLINE static void update_key_status (OPLL * opll)
{
  int ch;

  for (ch = 0; ch < 6; ch++)
    opll->slot_on_flag[ch * 2] = opll->slot_on_flag[ch * 2 + 1] = (opll->HiFreq[ch]) & 0x10;
}

/***********************************************************

                      Initializing

***********************************************************/

static void
OPLL_SLOT_reset (OPLL_SLOT * slot, int type)
{
  slot->type = type;
  slot->sintbl = waveform[0];
  slot->phase = 0;
  slot->dphase = 0;
  slot->output[0] = 0;
  slot->output[1] = 0;
  slot->feedback = 0;
  slot->eg_mode = SETTLE;
  slot->eg_phase = EG_DP_WIDTH;
  slot->eg_dphase = 0;
  slot->rks = 0;
  slot->tll = 0;
  slot->sustine = 0;
  slot->fnum = 0;
  slot->block = 0;
  slot->volume = 0;
  slot->pgout = 0;
  slot->egout = 0;
}

static void internal_refresh (void)
{
  makeDphaseTable ();
  makeDphaseARTable ();
  makeDphaseDRTable ();
  pm_dphase = (uint32) rate_adjust (PM_SPEED * PM_DP_WIDTH / (clk / 72));
  am_dphase = (uint32) rate_adjust (AM_SPEED * AM_DP_WIDTH / (clk / 72));
}

static void maketables (uint32 c, uint32 r)
{
  if(c != clk)
  {
    clk = c;
    makePmTable ();
    makeAmTable ();
    makeDB2LinTable ();
    makeAdjustTable ();
    makeTllTable ();
    makeRksTable ();
    makeSinTable ();
    //makeDefaultPatch ();
  }

  if(r != rate)
  {
    rate = r;
    internal_refresh ();
  }
}

OPLL *OPLL_new(uint32 clk,
	uint32 rate)
{
  OPLL *opll;

  maketables (clk, rate);

  if ((opll=(OPLL*)calloc(sizeof(OPLL),1))==NULL)
    return NULL;

  opll->mask = 0;

  OPLL_reset (opll);

  return opll;
}


void OPLL_delete(OPLL *opll)
{
  free (opll);
}

/* Reset whole of OPLL except patch datas. */
void OPLL_reset(OPLL *opll)
{
  int32 i;

  if(!opll)
    return;

  opll->adr = 0;
  opll->out = 0;

  opll->pm_phase = 0;
  opll->am_phase = 0;

  opll->mask = 0;

  for (i = 0; i < 12; i++)
    OPLL_SLOT_reset(&opll->slot[i], i%2);

  for (i = 0; i < 6; i++)
  {
    opll->key_status[i] = 0;
    //setPatch (opll, i, 0);
  }

  for (i = 0; i < 0x40; i++)
    OPLL_writeReg (opll, i, 0);

#ifndef EMU2413_COMPACTION
  opll->realstep = (uint32) ((1 << 31) / rate);
  opll->opllstep = (uint32) ((1 << 31) / (clk / 72));
  opll->oplltime = 0;
#endif
}

/* Force Refresh (When external program changes some parameters). */
void
OPLL_forceRefresh (OPLL * opll)
{
  int32 i;

  if(opll == NULL)
    return;

  for (i = 0; i < 12; i++)
  {
    UPDATE_PG (&opll->slot[i]);
    UPDATE_RKS (&opll->slot[i]);
    UPDATE_TLL (&opll->slot[i]);
    UPDATE_WF (&opll->slot[i]);
    UPDATE_EG (&opll->slot[i]);
  }
}

void
OPLL_set_rate (OPLL * opll, uint32 r)
{
  if(opll->quality)
    rate = 49716;
  else
    rate = r;
  internal_refresh ();
  rate = r;
}

void
OPLL_set_quality (OPLL * opll, uint32 q)
{
  opll->quality = q;
  OPLL_set_rate (opll, rate);
}

/*********************************************************

                 Generate wave data

*********************************************************/
/* Convert Amp(0 to EG_HEIGHT) to Phase(0 to 2PI). */
#if( SLOT_AMP_BITS - PG_BITS ) > 0
#define wave2_2pi(e)  ( (e) >> ( SLOT_AMP_BITS - PG_BITS ))
#else
#define wave2_2pi(e)  ( (e) << ( PG_BITS - SLOT_AMP_BITS ))
#endif

/* Convert Amp(0 to EG_HEIGHT) to Phase(0 to 4PI). */
#if( SLOT_AMP_BITS - PG_BITS - 1 ) == 0
#define wave2_4pi(e)  (e)
#elif( SLOT_AMP_BITS - PG_BITS - 1 ) > 0
#define wave2_4pi(e)  ( (e) >> ( SLOT_AMP_BITS - PG_BITS - 1 ))
#else
#define wave2_4pi(e)  ( (e) << ( 1 + PG_BITS - SLOT_AMP_BITS ))
#endif

/* Convert Amp(0 to EG_HEIGHT) to Phase(0 to 8PI). */
#if( SLOT_AMP_BITS - PG_BITS - 2 ) == 0
#define wave2_8pi(e)  (e)
#elif( SLOT_AMP_BITS - PG_BITS - 2 ) > 0
#define wave2_8pi(e)  ( (e) >> ( SLOT_AMP_BITS - PG_BITS - 2 ))
#else
#define wave2_8pi(e)  ( (e) << ( 2 + PG_BITS - SLOT_AMP_BITS ))
#endif



/* Update AM, PM unit */
static void
update_ampm (OPLL * opll)
{
  opll->pm_phase = (opll->pm_phase + pm_dphase) & (PM_DP_WIDTH - 1);
  opll->am_phase = (opll->am_phase + am_dphase) & (AM_DP_WIDTH - 1);
  opll->lfo_am = amtable[HIGHBITS (opll->am_phase, AM_DP_BITS - AM_PG_BITS)];
  opll->lfo_pm = pmtable[HIGHBITS (opll->pm_phase, PM_DP_BITS - PM_PG_BITS)];
}

/* PG */
INLINE static void
calc_phase (OPLL_SLOT * slot, int32 lfo)
{
  if(slot->patch.PM)
    slot->phase += (slot->dphase * lfo) >> PM_AMP_BITS;
  else
    slot->phase += slot->dphase;

  slot->phase &= (DP_WIDTH - 1);

  slot->pgout = HIGHBITS (slot->phase, DP_BASE_BITS);
}

/* EG */
static void
calc_envelope (OPLL_SLOT * slot, int32 lfo)
{
#define S2E(x) (SL2EG((int32)(x/SL_STEP))<<(EG_DP_BITS-EG_BITS))

  static uint32 SL[16] = {
    S2E (0.0), S2E (3.0), S2E (6.0), S2E (9.0), S2E (12.0), S2E (15.0), S2E (18.0), S2E (21.0),
    S2E (24.0), S2E (27.0), S2E (30.0), S2E (33.0), S2E (36.0), S2E (39.0), S2E (42.0), S2E (48.0)
  };

  uint32 egout;

  switch (slot->eg_mode)
  {

  case ATTACK:
    egout = AR_ADJUST_TABLE[HIGHBITS (slot->eg_phase, EG_DP_BITS - EG_BITS)];
    slot->eg_phase += slot->eg_dphase;
    if((EG_DP_WIDTH & slot->eg_phase)||(slot->patch.AR==15))
    {
      egout = 0;
      slot->eg_phase = 0;
      slot->eg_mode = DECAY;
      UPDATE_EG (slot);
    }
    break;

  case DECAY:
    egout = HIGHBITS (slot->eg_phase, EG_DP_BITS - EG_BITS);
    slot->eg_phase += slot->eg_dphase;
    if(slot->eg_phase >= SL[slot->patch.SL])
    {
      if(slot->patch.EG)
      {
        slot->eg_phase = SL[slot->patch.SL];
        slot->eg_mode = SUSHOLD;
        UPDATE_EG (slot);
      }
      else
      {
        slot->eg_phase = SL[slot->patch.SL];
        slot->eg_mode = SUSTINE;
        UPDATE_EG (slot);
      }
    }
    break;

  case SUSHOLD:
    egout = HIGHBITS (slot->eg_phase, EG_DP_BITS - EG_BITS);
    if(slot->patch.EG == 0)
    {
      slot->eg_mode = SUSTINE;
      UPDATE_EG (slot);
    }
    break;

  case SUSTINE:
  case RELEASE:
    egout = HIGHBITS (slot->eg_phase, EG_DP_BITS - EG_BITS);
    slot->eg_phase += slot->eg_dphase;
    if(egout >= (1 << EG_BITS))
    {
      slot->eg_mode = FINISH;
      egout = (1 << EG_BITS) - 1;
    }
    break;

  case FINISH:
    egout = (1 << EG_BITS) - 1;
    break;

  default:
    egout = (1 << EG_BITS) - 1;
    break;
  }

  if(slot->patch.AM)
    egout = EG2DB (egout + slot->tll) + lfo;
  else
    egout = EG2DB (egout + slot->tll);

  if(egout >= DB_MUTE)
    egout = DB_MUTE - 1;

  slot->egout = egout;
}

/* CARRIOR */
INLINE static int32
calc_slot_car (OPLL_SLOT * slot, int32 fm)
{
  slot->output[1] = slot->output[0];

  if(slot->egout >= (DB_MUTE - 1))
  {
    slot->output[0] = 0;
  }
  else
  {
    slot->output[0] = DB2LIN_TABLE[slot->sintbl[(slot->pgout+wave2_8pi(fm))&(PG_WIDTH-1)] + slot->egout];
  }

  return (slot->output[1] + slot->output[0]) >> 1;
}

/* MODULATOR */
INLINE static int32
calc_slot_mod (OPLL_SLOT * slot)
{
  int32 fm;

  slot->output[1] = slot->output[0];

  if(slot->egout >= (DB_MUTE - 1))
  {
    slot->output[0] = 0;
  }
  else if(slot->patch.FB != 0)
  {
    fm = wave2_4pi (slot->feedback) >> (7 - slot->patch.FB);
    slot->output[0] = DB2LIN_TABLE[slot->sintbl[(slot->pgout + fm)&(PG_WIDTH-1)] + slot->egout];
  }
  else
  {
    slot->output[0] = DB2LIN_TABLE[slot->sintbl[slot->pgout] + slot->egout];
  }

  slot->feedback = (slot->output[1] + slot->output[0]) >> 1;

  return slot->feedback;

}

static INLINE int16 calc (OPLL * opll)
{
  int32 inst = 0, out = 0;
  int32 i;

  update_ampm (opll);

  for (i = 0; i < 12; i++)
  {
    calc_phase(&opll->slot[i],opll->lfo_pm);
    calc_envelope(&opll->slot[i],opll->lfo_am);
  }

  for (i = 0; i < 6; i++)
    if(!(opll->mask & OPLL_MASK_CH (i)) && (CAR(opll,i)->eg_mode != FINISH))
      inst += calc_slot_car (CAR(opll,i), calc_slot_mod(MOD(opll,i)));

  out = inst;
  return (int16) out;
}

void OPLL_fill(OPLL* opll,
	int32 *buf,
	int32 c)//, int shift)
{
	if (c<=0)
		return;
	do
		*buf+= calc(opll)<<4;
	while(++buf,--c);
}

void OPLL_fillAbsolute(OPLL* opll,
	int16 *buf,
	int32 c)//, int shift)
{
	if (c<=0)
		return;
	do
		*buf+= (calc(opll)+32768)<<4;
	while(++buf,--c);
}

#ifdef EMU2413_COMPACTION
int16 OPLL_calc (OPLL * opll)
{
  return calc (opll);
}
#else
int16
OPLL_calc (OPLL * opll)
{
  if(!opll->quality)
   return calc (opll);

  while (opll->realstep > opll->oplltime)
  {
    opll->oplltime += opll->opllstep;
    opll->prev = opll->next;
    opll->next = calc (opll);
  }

  opll->oplltime -= opll->realstep;
  opll->out = (int16) (((double) opll->next * (opll->opllstep - opll->oplltime)
                          + (double) opll->prev * opll->oplltime) / opll->opllstep);

  return (int16) opll->out;
}
#endif

uint32
OPLL_setMask (OPLL * opll, uint32 mask)
{
  uint32 ret;

  if(opll)
  {
    ret = opll->mask;
    opll->mask = mask;
    return ret;
  }
  else
    return 0;
}

uint32
OPLL_toggleMask (OPLL * opll, uint32 mask)
{
  uint32 ret;

  if(opll)
  {
    ret = opll->mask;
    opll->mask ^= mask;
    return ret;
  }
  else
    return 0;
}

/****************************************************

                       I/O Ctrl

*****************************************************/

static void setInstrument(OPLL * opll, uint32 i, uint32 inst)
{
 const uint8 *src;
 OPLL_PATCH *modp, *carp;

 opll->patch_number[i]=inst;

 if(inst)
  src=default_inst[inst-1];
 else
  src=opll->CustInst;

 modp=&MOD(opll,i)->patch;
 carp=&CAR(opll,i)->patch;

 modp->AM=(src[0]>>7)&1;
 modp->PM=(src[0]>>6)&1;
 modp->EG=(src[0]>>5)&1;
 modp->KR=(src[0]>>4)&1;
 modp->ML=(src[0]&0xF);

 carp->AM=(src[1]>>7)&1;
 carp->PM=(src[1]>>6)&1;
 carp->EG=(src[1]>>5)&1;
 carp->KR=(src[1]>>4)&1;
 carp->ML=(src[1]&0xF);

 modp->KL=(src[2]>>6)&3;
 modp->TL=(src[2]&0x3F);

 carp->KL = (src[3] >> 6) & 3;
 carp->WF = (src[3] >> 4) & 1;

 modp->WF = (src[3] >> 3) & 1;

 modp->FB = (src[3]) & 7;

 modp->AR = (src[4]>>4)&0xF;
 modp->DR = (src[4]&0xF);

 carp->AR = (src[5]>>4)&0xF;
 carp->DR = (src[5]&0xF);

 modp->SL = (src[6]>>4)&0xF;
 modp->RR = (src[6]&0xF);

 carp->SL = (src[7]>>4)&0xF;
 carp->RR = (src[7]&0xF);
}


void
OPLL_writeReg (OPLL * opll, uint32 reg, uint32 data)
{

  int32 i, v, ch;

  data = data & 0xff;
  reg = reg & 0x3f;

  switch (reg)
  {
   case 0x00:
    opll->CustInst[0]=data;
    for (i = 0; i < 6; i++)
    {
      if(opll->patch_number[i] == 0)
      {
        setInstrument(opll, i, 0);
        UPDATE_PG (MOD(opll,i));
        UPDATE_RKS (MOD(opll,i));
        UPDATE_EG (MOD(opll,i));
      }
    }
    break;

  case 0x01:
    opll->CustInst[1]=data;
    for (i = 0; i < 6; i++)
    {
      if(opll->patch_number[i] == 0)
      {
        setInstrument(opll, i, 0);
        UPDATE_PG (CAR(opll,i));
        UPDATE_RKS (CAR(opll,i));
        UPDATE_EG (CAR(opll,i));
      }
    }
    break;

  case 0x02:
    opll->CustInst[2]=data;
    for (i = 0; i < 6; i++)
    {
      if(opll->patch_number[i] == 0)
      {
        setInstrument(opll, i, 0);
        UPDATE_TLL(MOD(opll,i));
      }
    }
    break;

  case 0x03:
    opll->CustInst[3]=data;
    for (i = 0; i < 6; i++)
    {
      if(opll->patch_number[i] == 0)
      {
        setInstrument(opll, i, 0);
        UPDATE_WF(MOD(opll,i));
        UPDATE_WF(CAR(opll,i));
      }
    }
    break;

  case 0x04:
    opll->CustInst[4]=data;
    for (i = 0; i < 6; i++)
    {
      if(opll->patch_number[i] == 0)
      {
        setInstrument(opll, i, 0);
        UPDATE_EG (MOD(opll,i));
      }
    }
    break;

  case 0x05:
    opll->CustInst[5]=data;
    for (i = 0; i < 6; i++)
    {
      if(opll->patch_number[i] == 0)
      {
        setInstrument(opll, i, 0);
        UPDATE_EG(CAR(opll,i));
      }
    }
    break;

  case 0x06:
    opll->CustInst[6]=data;
    for (i = 0; i < 6; i++)
    {
      if(opll->patch_number[i] == 0)
      {
        setInstrument(opll, i, 0);
        UPDATE_EG (MOD(opll,i));
      }
    }
    break;

  case 0x07:
    opll->CustInst[7]=data;
    for (i = 0; i < 6; i++)
    {
      if(opll->patch_number[i] == 0)
      {
        setInstrument(opll, i, 0);
         UPDATE_EG (CAR(opll,i));
      }
    }
    break;

  case 0x10:
  case 0x11:
  case 0x12:
  case 0x13:
  case 0x14:
  case 0x15:
    ch = reg - 0x10;
    opll->LowFreq[ch]=data;
    setFnumber (opll, ch, data + ((opll->HiFreq[ch] & 1) << 8));
    UPDATE_ALL (MOD(opll,ch));
    UPDATE_ALL (CAR(opll,ch));
    break;

  case 0x20:
  case 0x21:
  case 0x22:
  case 0x23:
  case 0x24:
  case 0x25:
    ch = reg - 0x20;
    opll->HiFreq[ch]=data;

    setFnumber (opll, ch, ((data & 1) << 8) + opll->LowFreq[ch]);
    setBlock (opll, ch, (data >> 1) & 7);
    setSustine (opll, ch, (data >> 5) & 1);
    if(data & 0x10)
      keyOn (opll, ch);
    else
      keyOff (opll, ch);
    UPDATE_ALL (MOD(opll,ch));
    UPDATE_ALL (CAR(opll,ch));
    update_key_status (opll);
    break;

  case 0x30:
  case 0x31:
  case 0x32:
  case 0x33:
  case 0x34:
  case 0x35:
    opll->InstVol[reg-0x30]=data;
    i = (data >> 4) & 15;
    v = data & 15;
    setInstrument(opll, reg-0x30, i);
    setVolume (opll, reg - 0x30, v << 2);
    UPDATE_ALL (MOD(opll,reg - 0x30));
    UPDATE_ALL (CAR(opll,reg - 0x30));
    break;

  default:
    break;

  }
}

void
OPLL_writeIO (OPLL * opll, uint32 adr, uint32 val)
{
  if(adr & 1)
    OPLL_writeReg (opll, opll->adr, val);
  else
    opll->adr = val;
}
