/*
 sound.c
  N2A0x PSG, pseuedo-Audio Processing Unit (pAPU) emulation

 * MacFCEU 0.8a3
 *
 * FCE Ultra - NES/Famicom Emulator
 *
 * Copyright notice for this file:
 *  sound.c Copyright (C) 2002 Xodnizel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <stdlib.h>

#include "driver.h"
#include "x6502.h"
#include "FCEU.h"
#include "WAVE.h"
#include "fsound.h"
#include "filter.h"
#include "state.h"

#include "utility/ftrig.h"
#include "memory.h"

#ifdef env68k
 #pragma segment N2A0x
#endif

#define EnvSpeedMask		0xF

/* rectangle: 4000-4007h */
#define RWaveLnMin		8
#define RWaveLnMax		0x7FF
#define RWaveLnHigh		0x700 //was 0xFF00, internally AND'd again
#define RSweepEnable		0x80
#define RSweepShift		4
#define REnvelopeLoop		0x20

/* triangle: 4008-400Bh */
#define TLoadMask		0x7F
#define TStartMask		0x80

/* noise: 400C-400Fh */
#define NShiftShort		0x80

/* 4010-4013h */
#define DMCFEnable		0x80
#define DMCFLoop		0x40
#define DMCFMask		0xF
#define DMCSIRQ			0x80

/* 4017h */
#define FMode5Step		0x80
#define FModeNoIRQ		0x40
#define FModeMask		0xC0

#define EORNoise(n,c)		(n=((n>>c^n>>14)&1)+n+n)

/* amplitude tables */
static int32	rectwlookup[32], trilookup[32];
/* 15*9+15*9+127 = 397 is adequate */
static int32	tpcmwlookup[398]; // triangle/noise/DMC -- increased for overflow, was 203, 400

/* virtual wave, prior to filtering */
int32		FCEUWave[12000]; // 2048+512 -- increased for higher sample rates, lower speeds
int32		FCEUWaveHi[40000];

/* post-filter results */
static int16	WaveFinal[12000]; //2048+512

EXPSOUND	GameExpSound; /* zeroed out in FCEU.c */

static uint8	TriCount;
static uint8	TriMode;
static int32	tristep;

static int32	wlcount[4];		/* Wave length counters. */

uint8		PSGReg[0x10];

uint8		IRQFrameMode;		/* $4017 xx000000 */
uint8		RawDALatch;		/* $4011 0xxxxxxx */

uint8		EnabledChannels;	/* 4015h */

typedef struct ENVUNIT/*(8)*/
{
	uint8	Speed;
	uint8	Mode;
	uint8	DecCountTo1;
	uint8	decvolume;
	int	reloaddec;
} ENVUNIT;

ENVUNIT		SndEnvUnits[3];

static int32	RectDutyCount[2];
uint8		RecSweepOn[2];
uint32		RecCurFreq[2];
static uint8	SweepCount[2];

static uint16	nreg;

uint8		Soundfcnt;
int32		Soundfhcnt; /* half-frame count */
int32		Soundfhinc; /* ...increment */

uint32		soundtsoffs;

/* Variables exclusively for low-quality sound. */
int32		nesincsize;
uint32		soundtsinc;
uint32		soundtsi;
static int32	sqacc[2];
/* LQ variables segment ends. */

int32		soundlencount[4];

static const uint8 lengthtable[0x20] =
{ /* updated to fit apu_ref.txt - 0x0D*2 (0x1A) on second row was 0x0e*2...was smoother */
	0x0A, 0xFE, 0x14, 0x02, 0x28, 0x04, 0x50, 0x06,
	0xA0, 0x08, 0x3C, 0x0A, 0x0E, 0x0C, 0x1A, 0x0E,
	0x0C, 0x10, 0x18, 0x12, 0x30, 0x14, 0x60, 0x16,
	0xC0, 0x18, 0x48, 0x1A, 0x10, 0x1C, 0x20, 0x1E
};

/* NTSC or Dendy */
static const uint16 GenNoiseFreqTable[0x10] =
{
	  4,   8,  16,  32,  64,  96, 128, 160,
	202, 254, 380, 508, 762,1016,2034,4068
};
static const uint16 GenDMCTable[0x10] =
{
	428, 380, 340, 320, 286, 254, 226, 214,
	190, 160, 142, 128, 106,  84,  72,  54
};

static const uint16 PALNoiseFreqTable[0x10] =
{
	  4,   7,  14,  30,  60,  88, 118, 148,
	188, 236, 354, 472, 708, 944,1890,3778
};
static const uint16 PALDMCTable[0x10] =
{ /* exactly 1/1.08 that of the NTSC table (rounded twice) */
	398, 354, 316, 298, 266, 236, 210, 198,
	176, 148, 131, 118,  98,  78,  66,  50,
};

/* one of the two tables above loaded by SetSoundVariables */
static uint32 CurrentNoiseFreqTable[0x10];
uint32		CurrentDMCTable[0x10];

// $4010        -        Frequency
// $4011        -        Actual data outputted
// $4012        -        Address register: $c000 + V*64
// $4013        -        Size register:  Size in bytes = (V+1)*64

int32		DMCacc;
int32		DMCPeriod;
uint8		DMCBitCount =-1; /* was 0/-8 */

uint8		DMCAddressLatch, DMCSizeLatch; /* writes to 4012 and 4013 */
uint8		DMCFormat;	/* Write to $4010 */

uint32		DMCAddress = 0;
int32		DMCCounter = 0; /* was DMCSize */

uint8		SIRQStat = 0;

uint8		DMCDMABuf = 0;
uint8		DMCBuffered = 0;
uint8		DMCActive = 0;

static void Dummyfunc(void) {};
void (*DoSQ1)(void)=Dummyfunc;
void (*DoSQ2)(void)=Dummyfunc;
void (*DoTriangle)(void)=Dummyfunc;
void (*DoNoise)(void)=Dummyfunc;
void (*DoPCM)(void)=Dummyfunc;

uint32	FCEUChannelBC[5];

static uint16	*ntable;

/* ---- macros */

#define LoadDMCPeriod(V)	(DMCPeriod=CurrentDMCTable[(uint8)(V)])

#define PrepDMC() \
	(DMCAddress=0x300+(uint32)DMCAddressLatch<<6,DMCCounter=((uint32)DMCSizeLatch<<4)+1)

/* Instantaneous?  Maybe the new freq value is being calculated all of the time... */
/* according to square_timer_div2, wavelengths have weight - probably calls for a completion */
/* Beware: this is now a macro */
#define CheckFreq(c,s)		(s&8 || !(((c>>(s&7))+c)&0x800))

//static void SQReload(int x,
//	uint8 V)
#define SQReload(x,V) do{ \
	if (EnabledChannels&(x+1)) /*(1<<x)*/ { \
		if (x) \
			DoSQ2(); \
		else \
			DoSQ1(); \
		soundlencount[x] = lengthtable[V>>3&31]; \
	} \
	RecSweepOn[x] = PSGReg[x*4+1]&RSweepEnable; \
	RecCurFreq[x] = PSGReg[x*4+2]|((uint32)V<<8&RWaveLnHigh); \
	SweepCount[x] = (PSGReg[x*4+1]>>RSweepShift&7)+1; \
	RectDutyCount[x] = 7; \
	SndEnvUnits[x].reloaddec = 1; \
}while(0)

/* ---- prototypes */

DECLFW(WritePSG0);
DECLFW(WritePSG1);
DECLFW(WritePSG2);
DECLFW(WritePSG3);
DECLFW(WritePSG4);
DECLFW(WritePSG5);
DECLFW(WritePSG6);
DECLFW(WritePSG7);
DECLFW(WritePSG);
DECLFW(WritePSGA);
DECLFW(WritePSGB);
DECLFW(WritePSGC);
DECLFW(WritePSGE);
DECLFW(WritePSGF);
DECLFW(WriteDMCCtl);
DECLFW(WriteDMCDA);
DECLFW(WriteDMCAddr);
DECLFW(WriteDMCSize);
DECLFW(SndStatusWrite);
DECLFR(SndStatusRead);
DECLFW(WriteIRQFM);

/* rect A */
DECLFW(WritePSG0)
{
(void)A;
	DoSQ1();
	SndEnvUnits[0].Mode = (V>>4)&3;
	SndEnvUnits[0].Speed = V&EnvSpeedMask;
	PSGReg[0] = V;
}

DECLFW(WritePSG1)
{
(void)A;
	RecSweepOn[0] = V&RSweepEnable;
	PSGReg[1] = V;
}

DECLFW(WritePSG2)
{
(void)A;
	DoSQ1();
	RecCurFreq[0]&= 0xFF00; //should wait until some completion
	RecCurFreq[0]|= V;
	PSGReg[2] = V;
}

DECLFW(WritePSG3)
{
(void)A;
	SQReload(0,V);
	PSGReg[3] = V;
}

/* rect B */
DECLFW(WritePSG4)
{
(void)A;
	DoSQ2();
	SndEnvUnits[1].Mode = (V>>4)&3;
	SndEnvUnits[1].Speed = V&EnvSpeedMask;
	PSGReg[4] = V;
}

DECLFW(WritePSG5)
{
(void)A;
	RecSweepOn[1] = V&RSweepEnable;
	PSGReg[5] = V;
}

DECLFW(WritePSG6)
{
(void)A;
	DoSQ2();
	RecCurFreq[1]&= 0xFF00;
	RecCurFreq[1]|= V;
	PSGReg[6] = V;
}

DECLFW(WritePSG7)
{
(void)A;
	SQReload(1,V);
	PSGReg[7] = V;
}

/* 8,9,D */
DECLFW(WritePSG)
{
	PSGReg[A-0x4000] = V;
}

/* triangle */
DECLFW(WritePSGA)
{
(void)A;
	DoTriangle();
	PSGReg[10] = V;
}

DECLFW(WritePSGB)
{
(void)A;
	DoTriangle();
	if (EnabledChannels&4)
		soundlencount[2] = lengthtable[(V>>3)&31];
	TriMode = 1; /* Load mode */
	PSGReg[11] = V;
}

DECLFW(WritePSGC)
{
(void)A;
	DoNoise();
	SndEnvUnits[2].Mode = V>>4&3;
	SndEnvUnits[2].Speed = V&EnvSpeedMask;
	PSGReg[12] = V;
}

DECLFW(WritePSGE)
{
(void)A;
	DoNoise();
	PSGReg[14] = V;
}

DECLFW(WritePSGF)
{
(void)A;
	DoNoise();
	if (EnabledChannels&8)
		soundlencount[3] = lengthtable[(V>>3)&31];
	SndEnvUnits[2].reloaddec = 1;
	PSGReg[15] = V;
}

DECLFW(WriteDMCCtl)
{
(void)A;
	DoPCM();
	LoadDMCPeriod(V&DMCFMask);
	DMCFormat = V;
	if (SIRQStat&DMCSIRQ)
		if (!(V&DMCFEnable))
		{
			XFrame.IRQlow&=~FCEU_IQDMC; //X6502_IRQEnd
			SIRQStat&=~0x80;
		} /*else -- this wasn't found
			XFrame.IRQlow|= FCEU_IQDMC; //X6502_IRQBegin
		*/
}

DECLFW(WriteDMCDA)
{
(void)A;
	DoPCM();
	RawDALatch = V&0x7F;
}

DECLFW(WriteDMCAddr)
{
(void)A;
	DMCAddressLatch = V;
}

DECLFW(WriteDMCSize)
{
(void)A;
	DMCSizeLatch = V;
}

void FASTAPASS(1) FrameSoundStuff(uint8 V);
void FrameSoundUpdate(void);
void RDoDMC(void);

#ifdef env68k
void DMCDMA(void);
#else
static
#endif
INLINE void DMCDMA(void)
{ /* rewritten as a separate function; 2011.02.12 */
//	X6502_DMR(DMCAddress); -- 3 DMR cyc in the correct form
	X6502_DMR(DMCAddress); //this cycle should be discounted if from a $4xxx addr
	X6502_DMR(DMCAddress);
	DMCDMABuf = X6502_DMR(DMCAddress);
	DMCAddress = (uint16)(DMCAddress+1)|0x8000;
	DMCBuffered = 1;

	if (!--DMCCounter)
		if (DMCFormat&DMCFLoop)
			PrepDMC();
		else if (DMCFormat&DMCFEnable) /* SIRQStat was prev. set anyway */
		{
			SIRQStat|= 0x80;
			XFrame.IRQlow|= FCEU_IQDMC; /*X6502_IRQBegin*/
		}
}

DECLFW(SndStatusWrite) //4015h
{ /* this should now be 100% accurate */
(void)A;
	DoSQ1();
	DoSQ2();
	DoTriangle();
	DoNoise();
	DoPCM();

	SIRQStat&=~0x80;
	XFrame.IRQlow&=~FCEU_IQDMC; /*X6502_IRQEnd*/
	EnabledChannels = V&=31;

	/* Force length counters to 0. */
	if (!(V&1))
		soundlencount[0] = 0;
	if (!(V&2))
		soundlencount[1] = 0;
	if (!(V&4))
		soundlencount[2] = 0;
	if (!(V&8))
		soundlencount[3] = 0;
	if (!(V&0x10))
		DMCCounter = 0;
	else if (!DMCCounter)
	{
		PrepDMC();
		if (!DMCBuffered) /* 2011.02.12 */
			DMCDMA();
	}
}

DECLFR(SndStatusRead) //4015h
{ /* other than top-precision timing, this is accurate */
	uint8		ret;
(void)A;
	ret = DMCCounter!=0, ret+= ret;
	ret+= soundlencount[3]!=0, ret+= ret;
	ret+= soundlencount[2]!=0, ret+= ret;
	ret+= soundlencount[1]!=0, ret+= ret;
	ret+= soundlencount[0]!=0;
	ret+= SIRQStat;

	if (!fceuindbg)
		SIRQStat&=~0x40,
			XFrame.IRQlow&=~FCEU_IQFCOUNT; /*X6502_IRQEnd*/

	return ret;
}

DECLFW(WriteIRQFM) //4017h
{ /* other than top-precision timing, this is now accurate */
(void)A; /* should round if on odd cycle */
	Soundfcnt = 0;
	if (V&FMode5Step)
		FrameSoundUpdate();
	Soundfcnt = 0x40;
	Soundfhcnt = Soundfhinc;
	if ((IRQFrameMode=V&FModeMask)==0)
		; /* increase latency */
	else if (V&FModeNoIRQ) /* added no-irq check; 2010.06.02 */
		XFrame.IRQlow&=~FCEU_IQFCOUNT, /*X6502_IRQEnd*/
			SIRQStat&=~0x40;
}

#define sweepIx(P) do{int32 mod; \
	if(!(PSGReg[P*4]&REnvelopeLoop) && soundlencount[P]>0) \
		--soundlencount[P]; \
	/* Frequency Sweep Code Here */ \
	/* xxxx 0000 */ \
	/* xxxx = hz.  120/(P+1) */ \
	if(!RecSweepOn[P]) \
		break; /* no sweep */ \
	if(SweepCount[P]>0) \
		--SweepCount[P]; \
	if(SweepCount[P]>0) \
		break; \
	SweepCount[P]=(PSGReg[P*4+1]>>RSweepShift&7)+1; /*+1;*/ \
	if(PSGReg[P*4+1]&8) { \
		if(RecCurFreq[P] && (mod=PSGReg[P*4+1]&7)!=0 /*&& RecSweepOn[P]&0x80*/ ) \
			RecCurFreq[P]-=(P^1)+(RecCurFreq[P]>>mod); \
	} else if(((mod=RecCurFreq[P]>>(PSGReg[P*4+1]&7))+RecCurFreq[P])&0x800) \
		RecSweepOn[P]=0,RecCurFreq[P]=0; \
	else if(RecCurFreq[P] && PSGReg[P*4+1]&7) /*&& RecSweepOn[P]&0x80*/ \
		RecCurFreq[P]+=mod; \
}while(0)

#define envelopeDecayIx(P) do{ \
	if(SndEnvUnits[P].reloaddec) { \
		SndEnvUnits[P].reloaddec=0,SndEnvUnits[P].decvolume=15; \
		SndEnvUnits[P].DecCountTo1=SndEnvUnits[P].Speed+1; \
		break; \
	} \
	if(SndEnvUnits[P].DecCountTo1>0) \
		--SndEnvUnits[P].DecCountTo1; \
	if(SndEnvUnits[P].DecCountTo1) \
		break; \
	SndEnvUnits[P].DecCountTo1=SndEnvUnits[P].Speed+1; \
	if(SndEnvUnits[P].decvolume || SndEnvUnits[P].Mode&2) \
		--SndEnvUnits[P].decvolume,SndEnvUnits[P].decvolume&=15; \
}while(0)

void FASTAPASS(1) FrameSoundStuff(uint8 V)
{
	DoSQ1();
	DoSQ2();
	DoNoise();
	DoTriangle();

	//!(V&1)
	if (!(V&0x40)) /* Envelope decay, linear counter, length counter, freq sweep */
	{
		if (!(PSGReg[8]&RSweepEnable) && soundlencount[2]>0)
			--soundlencount[2];
		if (!(PSGReg[12]&REnvelopeLoop) && soundlencount[3]>0)
			--soundlencount[3];
		sweepIx(0);
		sweepIx(1);
		sweepIx(2);
	}

	/* Now do envelope decay + linear counter. */
	if (TriMode) // In load mode?
		TriCount = PSGReg[8]&TLoadMask;
	else if (TriCount)
		--TriCount;
	if (!(PSGReg[8]&TStartMask))
		TriMode = 0;
	envelopeDecayIx(0);
	envelopeDecayIx(1);
	envelopeDecayIx(2);
}

#undef sweepIx
#undef envelopeDecayIx

/* a number of functions written into 2A0x.a */
void FrameSoundUpdate(void)
{
 // Linear counter:  Bit 0-6 of $4008
 // Length counter:  Bit 4-7 of $4003, $4007, $400b, $400f
	if (IRQFrameMode&FMode5Step)
	{
		if (Soundfcnt==0xC0) /*==3*/
			Soundfhcnt+= Soundfhinc;
	} else if (!(Soundfcnt|IRQFrameMode)) /* can't have both incr and IRQ */
		SIRQStat|= 0x40, XFrame.IRQlow|= FCEU_IQFCOUNT; /*X6502_IRQBegin*/
	FrameSoundStuff(Soundfcnt);
	Soundfcnt+= 0x40;
}

uint8		DMCShiftF[2] = {0}; /* first byte kept zero for BE 68K Asm */
#define DMCOutBuf DMCShiftF[1]

void FASTAPASS(1) FCEU_SoundCPUHook(int cycles)
{
	/*was `<=0` -- IRQ for 0 does occur, partial "clock_jitter" fix; 2011.02.10*/
	/* The only thing left is jitter for odd cycle frames */
	/* I see nothing in ChangeLog of issues regarding this line */
	if ((Soundfhcnt-=cycles*48)<0)
	{
		FrameSoundUpdate();
		Soundfhcnt+= Soundfhinc;
	}
	/* DMCDMA was prev. here */

	if ((DMCacc-=cycles)>0)
		return;
	do
	{
		if (DMCActive)
		{
			uint8	t = (DMCOutBuf<<2&4)-2;

			if (FSettings.SndRate)
			{ /* Unbelievably ugly hack */
				soundtsoffs+= DMCacc;
				DoPCM();
				soundtsoffs-= DMCacc;
			}
			if ((RawDALatch+=t)&0x80) /* overflow */
				RawDALatch-= t;
		}
		DMCOutBuf>>= 1;
		if (++DMCBitCount==0)
		{
			DMCBitCount =-8;
			//if (DMCHaveDMA) DMCHaveSample = 1; else...
			if ((DMCActive=DMCBuffered)!=0)
			{
				DMCBuffered = 0;
				DMCOutBuf = DMCDMABuf;
				if (DMCCounter) //&& !DMCBuffered)
					DMCDMA();
			}
		}
	} while((DMCacc+=DMCPeriod)<=0);
}

void RDoDMC(void)
{
	uint32		V;

	for(V=FCEUChannelBC[4]; V<SOUNDTS; ++V)
		FCEUWaveHi[V]+= (uint32)RawDALatch<<16;

	FCEUChannelBC[4] = SOUNDTS;
}

/* This has the correct phase.  Don't mess with it. */
static INLINE void RDoSQ(int x)
{
	int32		V;
	uint32		amp;
	int32		rthresh;
	int32		*D;
	int32		currdc;
	int32		cf;
	int32		rc;

	if (RecCurFreq[x]<RWaveLnMin || RecCurFreq[x]>RWaveLnMax)
		goto endit;
	cf = RecCurFreq[x], rc = PSGReg[x*4+1];
	if (!CheckFreq(cf,rc))
		goto endit;
	if (!soundlencount[x])
		goto endit;

	if (SndEnvUnits[x].Mode&1)
		amp = SndEnvUnits[x].Speed;
	else
		amp = SndEnvUnits[x].decvolume;
	amp<<= 25;

	//rthresh = RectDuties[PSGReg[x*4]>>6];
	rthresh = PSGReg[x*4]>>6;
	if ((rthresh+=rthresh)==0)
		rthresh = 1;

	D = &FCEUWaveHi[FCEUChannelBC[x]];
	V = SOUNDTS-FCEUChannelBC[x];

	currdc = RectDutyCount[x];
	cf = (RecCurFreq[x]+1)*2;
	rc = wlcount[x];

	if (V>0)
		do
		{
			if (currdc<rthresh)
				*D+= amp;
			if (!--rc)
				rc = cf, currdc = (currdc+1)&7;
		} while(++D,--V);

	RectDutyCount[x] = currdc;
	wlcount[x] = rc;
endit:	FCEUChannelBC[x] = SOUNDTS;
}

static void RDoSQ1(void)
{
	RDoSQ(0);
}

static void RDoSQ2(void)
{
	RDoSQ(1);
}

#define makettable(x) do{int _y,_z; \
	inie[x]=nesincsize; \
	_y=RecCurFreq[x],_z=PSGReg[x*4+1]; \
	if(RecCurFreq[x]<RWaveLnMin || RecCurFreq[x]>RWaveLnMax || !CheckFreq(_y,_z) \
			|| !soundlencount[x] || !(FSettings.soundactivech&(x+1))) \
		inie[x]=0; \
	if(SndEnvUnits[x].Mode&1) \
		amp[x]=SndEnvUnits[x].Speed; \
	else \
		amp[x]=SndEnvUnits[x].decvolume; \
	if(!inie[x]) amp[x]=0;	/* Correct? Buzzing in MM2, others otherwise... */ \
	_y=PSGReg[x*4]>>6; /*rthresh[x]=RectDuties[PSGReg[x*4]>>6];*/ \
	if((_y+=_y)==0) _y=1; \
	rthresh[x]=_y; \
	_y=0; \
	do \
		if(_y<rthresh[x]) \
			ttable[x][_y]=amp[x]; \
		else \
			ttable[x][_y]=0; \
	while(++_y<8); \
	freq[x]=RecCurFreq[x]+1<<1+17; \
} while(0)

static void RDoSQLQ(void)
{
	int32		V, end;
	int32		amp[2];
	int32		rthresh[2];
	int32		freq[2];
	int32		inie[2];
	int32		ttable[2][8];
	int32		totalout;

	V = FCEUChannelBC[0];
	if ((end=((unsigned)SOUNDTS<<16)/soundtsinc)<=V)
		return;
	FCEUChannelBC[0] = end;

	/* unrolled */
	makettable(0);
	makettable(1);

	totalout = rectwlookup[ ttable[0][RectDutyCount[0]] + ttable[1][RectDutyCount[1]] ];
	if (!(inie[0]|inie[1]))
	{
		do FCEUWave[V>>4]+=totalout; while(++V<end);
		RectDutyCount[0]&= 7, RectDutyCount[1]&= 7;
		return;
	}

	do
	{
		//int tmpamp=0;
		//if(RectDutyCount[0]<rthresh[0])
		// tmpamp=amp[0];
		//if(RectDutyCount[1]<rthresh[1])
		// tmpamp+=amp[1];
		//tmpamp=wlookup1[tmpamp];
		//tmpamp = wlookup1[ ttable[0][RectDutyCount[0]] + ttable[1][RectDutyCount[1]] ];

		FCEUWave[V>>4]+= totalout; //tmpamp;

		sqacc[1]-= inie[1];
		if ((sqacc[0]-=inie[0])<=0)
		{
			do
				++RectDutyCount[0];
			while((sqacc[0]+=freq[0])<=0);
			RectDutyCount[0]&= 7;

			if (sqacc[1]<=0)
				goto second; /* dodge unnecessary lookup; 2009.12.27 09:43 EST */
			totalout = rectwlookup[ ttable[0][RectDutyCount[0]]
				+ttable[1][RectDutyCount[1]] ];
			continue;
		}
		if (sqacc[1]<=0)
		{
		second:	do
				++RectDutyCount[1];
			while((sqacc[1]+=freq[1])<=0);
			RectDutyCount[1]&= 7;

			totalout = rectwlookup[ ttable[0][RectDutyCount[0]]
				+ttable[1][RectDutyCount[1]] ];
		}
	} while(++V<end);
}

static void RDoTriangle(void)
{
	uint32		V;
	int32		tfreq = (PSGReg[10]|((uint32)PSGReg[11]<<8&RWaveLnHigh))+1;
	uint32		tcout = trilookup[tristep&31];
	/* frequency check was missing; 2011.02.10 */
	if (tfreq<=4 || !soundlencount[2] || !TriCount)
	{ /* Counter is halted, but we still need to output. */
		int32	*start = &FCEUWaveHi[FCEUChannelBC[2]];
		int32	count;

		if ((count=SOUNDTS-FCEUChannelBC[2])!=0)
			do
				*start+++= tcout;
			while(--count);
		goto end;
	}

	for(V=FCEUChannelBC[2]; V<SOUNDTS; ++V)
	{
		FCEUWaveHi[V]+= tcout;
		if (!--wlcount[2])
		{
			wlcount[2] = tfreq;
			tcout = trilookup[++tristep&31];
		}
	}
end:	FCEUChannelBC[2] = SOUNDTS;
}

/* noise channel's always the slowest (optimize as much as possible) */
static void RDoTriangleNoiseDMCLQ(void)
{
	static uint32	tcout = 0;
	static int32	triacc = 0;
	static int32	noiseacc = 0;

	int32		V, end;
	uint32		tfreq, tlen, nlen;
	uint32		namptab[2]; /* namptab[1] is zero */
	uint32		noiseout, dmcout;
	int32		totalout;

	V = FCEUChannelBC[2];
	if ((end=((uint32)SOUNDTS<<16)/soundtsinc)<=V)
		return;
	FCEUChannelBC[2] = end;

	tlen = nlen = nesincsize;

	if (!(FSettings.soundactivech&0x10))
		dmcout = 0;
	else
		dmcout = RawDALatch;

	if ((tfreq=(PSGReg[10]|((uint32)PSGReg[11]<<8&RWaveLnHigh))+1)<=4
			|| !soundlencount[2] || !TriCount
			|| !(FSettings.soundactivech&4))
		tlen = 0;
	tfreq<<= 16+1;

	namptab[1] = 0;
	if (!soundlencount[3] || !(FSettings.soundactivech&8))
		namptab[0] = nlen = 0; /* Quick hack speedup, set [nlen] to 0 */
	else if (SndEnvUnits[2].Mode&1)
		namptab[0] = FSettings.soundnscale*SndEnvUnits[2].Speed;
	else
		namptab[0] = FSettings.soundnscale*SndEnvUnits[2].decvolume;

	noiseout = namptab[nreg>>14&1];
	//tcout = trilookup[tristep&31];
	totalout = tpcmwlookup[tcout+noiseout+dmcout];

	if (tlen && nlen)
	{
		uint32	nr = nreg, tc = tcout;

		do
		{
			FCEUWave[V>>4]+= totalout;
			noiseacc-= nlen;
			triacc-= tlen;
			if (triacc<=0)
			{
				do
					++tristep;
				while((triacc+=tfreq)<=0);
				tc = trilookup[tristep&31];
				if (noiseacc>0)
					goto samp;
				goto nsamp;
			}
			if (noiseacc<=0)
			{
				const uint16	*nt;

				int	nfreq;
			nsamp:
				nfreq = CurrentNoiseFreqTable[PSGReg[14]&15]<<16+1;
				nt = PSGReg[14]&NShiftShort ? ntable:ntable+32768;
				do
					nr = nt[nr];
				while((noiseacc+=nfreq)<=0);
				noiseout = namptab[nr>>14];
			samp:	totalout = tpcmwlookup[tc+noiseout+dmcout];
			}
		} while(++V<end);
		tcout = tc, nreg = nr;
	} else if (tlen)
	{
		uint32	tc = tcout;

		do
		{
			FCEUWave[V>>4]+= totalout;
			if ((triacc-=tlen)<=0)
			{
				do
					++tristep;
				while((triacc+=tfreq)<=0);
				tc = trilookup[tristep&31];
				totalout = tpcmwlookup[tc+noiseout+dmcout];
			}
		} while(++V<end);
		tcout = tc;
	} else if (nlen)
	{
		uint32	nr = nreg;

		do
		{
			FCEUWave[V>>4]+= totalout;
			noiseacc-= nlen;
			if (noiseacc<=0)
			{
				const uint16	*nt;

				int	nfreq;

				nfreq = CurrentNoiseFreqTable[PSGReg[14]&15]<<16+1;
				nt = PSGReg[14]&NShiftShort ? ntable:ntable+32768;
				do
					nr = nt[nr];
				while((noiseacc+=nfreq)<=0);
				noiseout = namptab[nr>>14];
				totalout = tpcmwlookup[tcout+noiseout+dmcout];
			}
		} while(++V<end);
		nreg = nr;
	} else
		do
			FCEUWave[V>>4]+= totalout;
		while(++V<end);
}

static void RDoNoise(void)
{
	int32		V, nfreq = CurrentNoiseFreqTable[PSGReg[14]&15];
	int32		outo;
	uint32		namptab[2];
	uint16		nr = nreg;

	if (SndEnvUnits[2].Mode&1)
		namptab[0] = (uint32)(FSettings.soundnscale
			*SndEnvUnits[2].Speed)<<16;
	else
		namptab[0] = (uint32)(FSettings.soundnscale
			*SndEnvUnits[2].decvolume)<<16;
	namptab[1] = 0;

	outo = namptab[nr&1]; //(nreg>>0xe)&1];

	if (!soundlencount[3] || !(FSettings.soundactivech&8))
		outo = namptab[0] = 0;

	if (PSGReg[14]&NShiftShort)
		for(V=FCEUChannelBC[3]; V<SOUNDTS; ++V)
		{
			FCEUWaveHi[V]+= outo;
			if (--wlcount[3]==0)
				wlcount[3] = nfreq,
					outo = namptab[(nr=ntable[nr])>>14];
		}
	else
		for(V=FCEUChannelBC[3]; V<SOUNDTS; ++V)
		{
			FCEUWaveHi[V]+= outo;
			if (--wlcount[3]==0)
				wlcount[3] = nfreq,
					outo = namptab[(nr=ntable[nr+32768])>>14];
		}
	nreg = nr;
	FCEUChannelBC[3] = SOUNDTS;
}

void SetNESSoundMap(void)
{
	SetWriteHandler(0x4000,0x4000,WritePSG0);
	SetWriteHandler(0x4001,0x4001,WritePSG1);
	SetWriteHandler(0x4002,0x4002,WritePSG2);
	SetWriteHandler(0x4003,0x4003,WritePSG3);
	SetWriteHandler(0x4004,0x4004,WritePSG4);
	SetWriteHandler(0x4005,0x4005,WritePSG5);
	SetWriteHandler(0x4006,0x4006,WritePSG6);
	SetWriteHandler(0x4007,0x4007,WritePSG7);
	SetWriteHandler(0x4008,0x400D,WritePSG);
	SetWriteHandler(0x400A,0x400A,WritePSGA);
	SetWriteHandler(0x400B,0x400B,WritePSGB);
	SetWriteHandler(0x400C,0x400C,WritePSGC);
	SetWriteHandler(0x400E,0x400E,WritePSGE);
	SetWriteHandler(0x400F,0x400F,WritePSGF);
	SetWriteHandler(0x4010,0x4010,WriteDMCCtl);
	SetWriteHandler(0x4011,0x4011,WriteDMCDA);
	SetWriteHandler(0x4012,0x4012,WriteDMCAddr);
	SetWriteHandler(0x4013,0x4013,WriteDMCSize);
	SetReadHandler(0x4015,0x4015,SndStatusRead);
	SetWriteHandler(0x4015,0x4015,SndStatusWrite);
	SetWriteHandler(0x4017,0x4017,WriteIRQFM);
}

static int32	inbuf = 0;

int FlushEmulateSound(void)
{
	uint32		b;
	int32		x, end, left;

	if ((x=timestamp)==0)
		return 0;
	if (!FSettings.SndRate)
	{
		left = end = 0;
		goto nosnd;
	}

	DoSQ1();
	DoSQ2();
	DoTriangle();
	DoNoise();
	DoPCM();

	if (FSettings.soundq)
	{
		int32	*tmpo = &FCEUWaveHi[soundtsoffs];

		if (FSettings.soundactivech&0x20)
			for(b=0; b<=SCSFME; ++b)
				if (FSettings.soundactivech&(64u<<b) && GameExpSound.HiFill[b])
					GameExpSound.HiFill[b]();
		/* ==0 already checked */
		do
			b = *tmpo, *tmpo++ =
				(b&0xFFFF)+tpcmwlookup[b>>16&0x1FF]+rectwlookup[b>>25];
		while(--x);
		end = NeoFilterSound(FCEUWaveHi,WaveFinal,SOUNDTS,&left);

		memmove(FCEUWaveHi,FCEUWaveHi+SOUNDTS-left,left*sizeof(uint32));
		memset(FCEUWaveHi+left,0,sizeof(FCEUWaveHi)-left*sizeof(uint32));

		for(b=0; b<=SCSFME; ++b)
			if (GameExpSound.HiSync[b])
				GameExpSound.HiSync[b](left);

		FCEUChannelBC[0] = FCEUChannelBC[1] = FCEUChannelBC[2] =
			FCEUChannelBC[3] = FCEUChannelBC[4] = left;
	} else
	{
		end = ((uint32)SOUNDTS<<16)/soundtsinc;
		if (FSettings.soundactivech&0x20)
			for(b=0; b<=SCSFME; ++b)
				if (FSettings.soundactivech&(64u<<b) && GameExpSound.Fill[b])
					GameExpSound.Fill[b](end&15);
		SexyFilter(FCEUWave,WaveFinal,end>>4);

		//if(FSettings.lowpass)
		// SexyFilter2(WaveFinal,end>>4);

		if (end&15) /* roll ending sample to beginning */
			FCEUWave[0] = FCEUWave[end>>4];
		FCEUWave[end>>4] = 0;
	}
nosnd:	if (FSettings.soundq)
		soundtsoffs = left;
	else
	{
		FCEUChannelBC[0] = FCEUChannelBC[1] = FCEUChannelBC[2] =
			FCEUChannelBC[3] = FCEUChannelBC[4] = end&15;
		soundtsoffs = (soundtsinc*(end&15))>>16, end>>= 4;
	}
	inbuf = end;

	FCEU_WriteWaveData(WaveFinal,end); /* does nothing if sound recording is off */
	return end;
}

int GetSoundBuffer(int16 **W)
{
	*W = WaveFinal;
	return inbuf;
}

/* FIXME:  Find out what sound registers get reset on reset.  I know $4001/$4005 don't,
due to that whole MegaMan 2 Game Genie thing.
*/

void FCEUSND_Reset(void)
{ /* mode is not initialize when fully powered */
	IRQFrameMode = 0; /* enabled - Dendy would do FModeNoIRQ */
	Soundfhcnt = Soundfhinc;
	Soundfcnt = 0;

	nreg = 1;

	wlcount[0] = wlcount[1] = 2048;
	if (nesincsize) /* lq mode */
		sqacc[0] = sqacc[1] = (2048UL<<17)/nesincsize;
	else
		sqacc[0] = sqacc[1] = 1;
	RecSweepOn[0] = RecSweepOn[1] = 0;
	RecCurFreq[0] = RecCurFreq[1] = 0;

	wlcount[2] = 1;	//2048;
	wlcount[3] = 2048;
	//some of these are cleared only on Power
	DMCAddressLatch = DMCSizeLatch = DMCFormat = DMCOutBuf = 0;
	DMCAddress = DMCCounter = DMCBuffered = DMCActive = 0;
	DMCacc = CurrentDMCTable[0]; //1; 2011.02.13
	SIRQStat = 0x00;

	RawDALatch = 0x00;
	TriCount = 0;
//	TriMode = 0;
	tristep = 7; //0;
	EnabledChannels = 0;

	soundlencount[0] = soundlencount[1] = soundlencount[2] = soundlencount[3] = 0;

	/* added; 2011.02.10 */
	WritePSG(3,0);
	WritePSG(7,0);
	SndEnvUnits[0].Mode = SndEnvUnits[1].Mode = SndEnvUnits[2].Mode =
		TriMode = SndEnvUnits[0].reloaddec = SndEnvUnits[1].reloaddec =
		SndEnvUnits[2].reloaddec = 1;
}

void FCEUSND_Power(void)
{
	SetNESSoundMap();
	FCEU_memsetl(FCEUWave,0,sizeof(FCEUWave));
	FCEU_memsetl(FCEUWaveHi,0,sizeof(FCEUWaveHi));
	FCEU_memsetl(SndEnvUnits,0,sizeof(SndEnvUnits));
	FCEUSND_Reset();

	/* added (actual may be in Reset, but this gives an option); 2009.12.27 */
	DMCBitCount = 0;

	{
		int		x;

		x = 4, FCEUChannelBC[4] = 0;
		do
			FCEUChannelBC[x-1] = ((int32*)PSGReg)[x-1] = 0;
		while(--x);
	}
	soundtsoffs = 0;
	LoadDMCPeriod(DMCFormat&DMCFMask);
}

static void SndSetFuncs(uint8 bits)
{
	if (FSettings.soundq)/*>=1*/
	{
		DoSQ1 = bits&1 ? RDoSQ1:Dummyfunc;
		DoSQ2 = bits&2 ? RDoSQ2:Dummyfunc;
		DoTriangle = bits&4 ? RDoTriangle:Dummyfunc;
		DoNoise = bits&8 ? RDoNoise:Dummyfunc;
		DoPCM = bits&0x10 ? RDoDMC:Dummyfunc;
	} else
	{
		DoSQ1 = bits&1 ? RDoSQLQ:Dummyfunc;
		DoSQ2 = bits&2 ? RDoSQLQ:Dummyfunc;
		DoTriangle = bits&4 ? RDoTriangleNoiseDMCLQ:Dummyfunc;
		DoNoise = bits&8 ? RDoTriangleNoiseDMCLQ:Dummyfunc;
		DoPCM = bits&0x10 ? RDoTriangleNoiseDMCLQ:Dummyfunc;
	}
}

/* The console 2A07 PAL CPU apparently operates at 1.6MHz */

#define ADJUSTED_PAL_CPU_RATIO		(double)(PAL_CPU/1600000)

static void LoadLUTs(void)
{
	const uint16	*nc, *dc;
	uint32		*na = CurrentNoiseFreqTable, *da = CurrentDMCTable;
	int		x;

	if (PAL)
		nc = PALNoiseFreqTable, dc = PALDMCTable;
	else
		nc = GenNoiseFreqTable, dc = GenDMCTable;
	x = 16;
	do
	{
		*na++ = *nc++;
		if (RateFixFlags&0x40) /* now optional; 2013.05.15 */
			*da++ = *dc++;
		else
			*da++ = umuldiv(*dc++,EmulationSpeedRatio,500000000)+1>>1;
	} while(--x);
}

static int BuildWLUTs(void)
{
	double		numer;
	int		x;
	/* speed boost: noise table */
	if (!ntable && (ntable=malloc(32768*2*sizeof(uint16)))==NULL)
	{ /* no memory for sound, for that matter... */
		FSettings.SndRate = 0;
		return 0;
	}
	x = 0;
	do
	{
		ntable[x] = (((x>>8^x>>14)&1)+x+x)&0x7FFF;
		ntable[x+32768] = (((x>>13^x>>14)&1)+x+x)&0x7FFF;
	} while(++x<32768);

	{ /* triangle scale table */
		uint32	n = FSettings.soundtscale;

		if (FSettings.soundq)
			n<<= 16;
		x = 0;
		do
			trilookup[x] = ((x&0x10 ? x:~x)&15)*n;
		while(++x<sizeof(trilookup)/sizeof(trilookup[0]));
	}

	/* rectangle filter */
	x = 1;
	numer = (95.52*64)*FSettings.soundrectamp;
	do
		rectwlookup[x] = numer/(8128.0/x+100);
	while(++x<32);
	/* triangle/noise/DMC filter */
	x = 1, numer = (163.67*64)*FSettings.soundtndamp;
	do
		tpcmwlookup[x] = numer/(24329.0/x+100);
	while(++x<sizeof(tpcmwlookup)/sizeof(tpcmwlookup[0]));

	return 1;
}

void SetSoundVariables(void)
{
	uint64		ratio;
	uint32		Rate;
	int		x;

	if (!EmulationSpeedRatio)
		EmulationSpeedRatio = 1000000000UL;
	LoadLUTs();

	rectwlookup[0] = tpcmwlookup[0] = 0;

	if ((Rate=FSettings.SndRate*16)!=0) /* *16 for LQ anti-aliasing */
	{ /* build the wave amplitude LUTs */
		if (!BuildWLUTs())
			goto silence;
		SndSetFuncs(FSettings.soundactivech);
	} else
	{
	silence:
		SndSetFuncs(0);
		return;
	}

	MakeFilters(Rate>>4);

	for(x=0; x<=SCSFME; ++x)
		if (GameExpSound.RChange[x])
			GameExpSound.RChange[x]();

	if (PAL)
	{
		_int64xset(ratio,(1L<<17)*(double)PAL_CPU);
		Soundfhinc = 16626*24;
	} else
	{
		_int64xset(ratio,(1L<<17)*(double)NTSC_CPU);
		Soundfhinc = 14915*24;
	}

	nesincsize = uint64divu32(ratio,Rate);

	sqacc[0] = sqacc[1] = 0;
	FCEU_memsetl(FCEUChannelBC,0,sizeof(FCEUChannelBC));

	LoadDMCPeriod(DMCFormat&DMCFMask); /* in case of changing from PAL to NTSC */

	/* There are 2 modes to this, and 4+2 NTSC stepping lengths
		normal-0	14912	14916
		linear-1	14916	14912
		normal-2	1	14916
		linear-3	1	14912
		other/0-4	1	14908
		other/0-5	14914	n/a
	*/
	soundtsinc = (unsigned)nesincsize;
	if (EmulationSpeedRatio!=1000000000UL)
		soundtsinc = umuldiv(soundtsinc,EmulationSpeedRatio,1000000000UL);
	soundtsinc>>= 1;
}

void FCEUI_Sound(uint32 Rate)
{
	if ((FSettings.OrgRate=Rate)!=0) /* original passed rate */
	{
		FSettings.SndRateF = (94u<<23)/Rate;
		if (!PAL && RateFixFlags&0x80) /* 2011.03.30 */
			Rate = ufract30mul(Rate,NTSC_CPU/1600000*1073741824.0+.5);
		else if (PAL && RateFixFlags&1) /* 2013.05.05 */
			Rate = ufract30mul(Rate,PAL_CPU/1600000*1073741824.0+.5);
	}
	FSettings.SndRate = Rate; /* emulated rate; may be adjusted */
	SetSoundVariables();
}

/* added; 2008.09.12 */
void FCEUI_SetSoundActiveChannels(uint16 bits)
{ /* flush before setting */
	DoSQ1();
	DoSQ2();
	DoTriangle();
	DoNoise();
	DoPCM();

	FSettings.soundactivech = bits;

	if (FSettings.SndRate)
		SndSetFuncs(bits);
}

void FCEUI_SetLowPass(int q) /* high-quality function */
{
	FSettings.lowpass = q;
}

void FCEUI_SetSoundQuality(int qty)
{
	FSettings.soundq = (unsigned)qty>255 ? 255:qty;
	SetSoundVariables();
}

void FCEUI_SetSoundVolume(int32 vol)
{
	FSettings.SndVolF = (int32)((uint32)(FSettings.SoundVolume=vol)
		<<(FSettings.soundq ? 15:15-4))/25;
}

/* added; 2010.03.23 */
void FCEUI_SetRectFilterAmp(int16 amp)
{
	FSettings.soundrectamp = amp;
	SetSoundVariables();
}

/* added; 2013.05.04 */
void FCEUI_SetTNDFilterAmp(int16 amp)
{
	FSettings.soundtndamp = amp;
	BuildWLUTs();
}

/* added; 2010.07.17 */
void FCEUI_SetNoiseScale(uint16 scale)
{ /* internal multiplier (no LUT) */
	FSettings.soundnscale = scale;
}

/* added; 2013.05.05 */
void FCEUI_SetTriangleScale(uint16 scale)
{
	FSettings.soundtscale = scale;
	BuildWLUTs();
}

SFORMAT FCEUSND_STATEINFO[] =
{
	{&Soundfhcnt,4|FCEUSTATE_RLSB,"FHCN"},
	{&Soundfcnt,1,"FCNT"},
	{ PSGReg,0x10,"PSG"},
	{&EnabledChannels,1,"ENCH"},
	{&IRQFrameMode,1,"IQFM"},
	{&nreg,2|FCEUSTATE_RLSB,"NREG"},
	{&TriMode,1,"TRIM"},
	{&TriCount,1,"TRIC"},

	{&SndEnvUnits[0].Speed,1,"E0SP"},
	{&SndEnvUnits[1].Speed,1,"E1SP"},
	{&SndEnvUnits[2].Speed,1,"E2SP"},

	{&SndEnvUnits[0].Mode,1,"E0MO"},
	{&SndEnvUnits[1].Mode,1,"E1MO"},
	{&SndEnvUnits[2].Mode,1,"E2MO"},

	{&SndEnvUnits[0].DecCountTo1,1,"E0D1"},
	{&SndEnvUnits[1].DecCountTo1,1,"E1D1"},
	{&SndEnvUnits[2].DecCountTo1,1,"E2D1"},

	{&SndEnvUnits[0].decvolume,1,"E0DV"},
	{&SndEnvUnits[1].decvolume,1,"E1DV"},
	{&SndEnvUnits[2].decvolume,1,"E2DV"},

	{&soundlencount[0],4|FCEUSTATE_RLSB,"LEN0"},
	{&soundlencount[1],4|FCEUSTATE_RLSB,"LEN1"},
	{&soundlencount[2],4|FCEUSTATE_RLSB,"LEN2"},
	{&soundlencount[3],4|FCEUSTATE_RLSB,"LEN3"},
	{ RecSweepOn,2,"SWEE"},
	{&RecCurFreq[0],4|FCEUSTATE_RLSB,"CRF1"},
	{&RecCurFreq[1],4|FCEUSTATE_RLSB,"CRF2"},
	{ SweepCount,2,"SWCT"},

	{&SIRQStat,1,"SIRQ"},

	{&DMCacc,4|FCEUSTATE_RLSB,"5ACC"},
	{&DMCBitCount,1,"5BIT"},
	{&DMCAddress,4|FCEUSTATE_RLSB,"5ADD"},
	{&DMCCounter,4|FCEUSTATE_RLSB,"5SIZ"},
	{&DMCShiftF,2,"5SHF"},

	{&DMCBuffered,1,"5HVD"}, //truncated from 5HVDM, 5HVSP
	{&DMCActive,1,"5HVS"},

	{&DMCSizeLatch,1,"5SZL"},
	{&DMCAddressLatch,1,"5ADL"},
	{&DMCFormat,1,"5FMT"},
	{&RawDALatch,1,"RWDA"},
	{0}
};

void FCEUSND_SaveState()
{
}

void FCEUSND_LoadState(int v)
{
	if (v<9819) /* these two prev. downshifted 6 */
		IRQFrameMode<<= 6, Soundfcnt<<= 6;
	PSGReg[1]&= 7, PSGReg[5]&= 7;
	DMCBitCount = (DMCBitCount&7)-8;
	LoadDMCPeriod(DMCFormat&DMCFMask);
	DMCAddress = (uint16)DMCAddress|0x8000;
	RawDALatch&= 0x7F;
	DMCShiftF[0] = 0; //for an Asm optimization
}
