/*
 ftrig.c
  integer fraction routines

  Created: 2008.07.15
  Revised: 2013.05.15

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "../enviro64.h"
#include "ftrig.h"

#ifdef env68k
 #pragma segment Lib
#else
ifixed vfix16mul(ifixed a,
	ifixed b)
{
	int64		v;

	(void)muls64s32(v,a,b);
	if (int64n(v))
	{
		(void)int64add32(v,-0x8000);
		int64shr(v,16);
		if (((a=int64lo(v))>>31)!=int64hi(v))
			return (ifixed)0x80000000; /* -overflow */
	} else
	{
		(void)int64add32(v,0x8000);
		int64shr(v,16);
		if (((a=int64lo(v))>>31)!=int64hi(v))
			return 0x7FFFFFFF; /* +overflow */
	}

	return a;
}

ifixed vfix16div(ifixed a,
	ifixed b)
{
	int64		v;

	if (b<0)
		int64sethl(v,0,-b>>1); /* round by -=denom/2 */
	else if (!b)
		if (a<0)
			return (ifixed)0x80000000;
		else
			return 0x7FFFFFFF;
	else
		int64sethl(v,0,b>>1); /* round by +=denom/2 */
	if (a<0) /* round currently positive; copy sign */
		(void)int64neg(v);
	(void)int64addhl(v,a>>16,(unsigned)a<<16);
	(void)int64divs32(v,b);

	return int64toi(v);
}

/* square root half (16 bit) table */
static const uint16	srhlut[] =
{
	0x8084,0x8180,0x827C,0x8378,0x8470,0x8568,0x865C,0x8750,
	0x8840,0x8930,0x8A1C,0x8B0C,0x8BF4,0x8CE0,0x8DC8,0x8EAC,
	0x8F94,0x9074,0x9158,0x9238,0x9318,0x93F8,0x94D4,0x95B0,
	0x9688,0x9760,0x9838,0x9910,0x99E8,0x9ABC,0x9B8C,0x9C60,
	0x9D30,0x9E00,0x9ED0,0x9F9C,0xA06C,0xA138,0xA200,0xA2CC,
	0xA394,0xA45C,0xA524,0xA5E8,0xA6AC,0xA770,0xA834,0xA8F8,
	0xA9B8,0xAA78,0xAB38,0xABF8,0xACB8,0xAD74,0xAE30,0xAEEC,
	0xAFA8,0xB060,0xB11C,0xB1D4,0xB28C,0xB344,0xB3F8,0xB4B0,
	0xB564,0xB618,0xB6CC,0xB780,0xB830,0xB8E4,0xB994,0xBA44,
	0xBAF4,0xBBA0,0xBC50,0xBCFC,0xBDAC,0xBE58,0xBF04,0xBFB0,
	0xC058,0xC104,0xC1AC,0xC254,0xC300,0xC3A4,0xC44C,0xC4F4,
	0xC598,0xC640,0xC6E4,0xC788,0xC82C,0xC8D0,0xC974,0xCA14,
	0xCAB8,0xCB58,0xCBF8,0xCC9C,0xCD3C,0xCDD8,0xCE78,0xCF18,
	0xCFB4,0xD054,0xD0F0,0xD18C,0xD228,0xD2C4,0xD360,0xD3FC,
	0xD494,0xD530,0xD5C8,0xD660,0xD6F8,0xD794,0xD828,0xD8C0,
	0xD958,0xD9F0,0xDA84,0xDB1C,0xDBB0,0xDC44,0xDCD8,0xDD6C,
	0xDE00,0xDE94,0xDF28,0xDFBC,0xE04C,0xE0E0,0xE170,0xE200,
	0xE294,0xE324,0xE3B4,0xE444,0xE4D4,0xE560,0xE5F0,0xE680,
	0xE70C,0xE798,0xE828,0xE8B4,0xE940,0xE9CC,0xEA58,0xEAE4,
	0xEB70,0xEBFC,0xEC88,0xED10,0xED9C,0xEE24,0xEEAC,0xEF38,
	0xEFC0,0xF048,0xF0D0,0xF158,0xF1E0,0xF268,0xF2F0,0xF374,
	0xF3FC,0xF480,0xF508,0xF58C,0xF614,0xF698,0xF71C,0xF7A0,
	0xF824,0xF8A8,0xF92C,0xF9B0,0xFA34,0xFAB4,0xFB38,0xFBBC,
	0xFC3C,0xFCC0,0xFD40,0xFDC0,0xFE44,0xFEC4,0xFF44,0xFFFF
};


uint32 usqrt64(uint64 n)
{
	uint64		t;
	/* get the highest ordered bit, and shift up... */
	uint32		r, s, a, b = int64lo(n);

	if ((a=int64hi(n))==0)
		return usqrt(b);

	s = cntlz32(a)&-2;
	a = a<<s|(b>>=32-s);
	r = srhlut[(a>>24)-64];

	s>>= 1; /* shift it back down by half when we are finished */
	r+= a/r; r>>= 1;
	r<<= 16-s; /* reset scale */

	/* divide and average */
	t = n; (void)uint64divu32(t,r); (void)uint64addu32(t,r); uint64half(t);

	if (int64hi(t))
		r = 0xFFFFFFFFUL; /* overflow */
	else
		r = int64lo(t);
	/* to prevent pre-rounding, we subtract one and test for rounding twice */
	(void)mulu64u32(t,r,r-1);
	(void)uint64sub(t,n);
	if (!int64n(t))
		--r;
	if (r==0xFFFFFFFFUL)
		return 0xFFFFFFFFUL;
	(void)mulu64u32(t,r,r+1);
	(void)uint64sub(t,n);
	if (int64n(t))
		++r;
	return r;
}

uint32 usqrt(uint32 n)
{
	/* to do this effectively, we get the highest ordered bit, and shift up... */
	uint32		a, r, s = cntlz32(n)&-2;

	if ((a=n<<s)==0)
		return 0;
	r = srhlut[(a>>24)-64];

	s>>= 1; /* shift it back down by half when finished */
	/* the table is large enough that we only have to divide and average once */
	r+= a/r; r>>= 1;
	r>>= s; /* reset scale */

	/* to prevent pre-rounding, subtract one and test for rounding twice */
	--r;
	if (r*r+r<n)
		++r;
	if ((signed)(r*r+r-n)<0)
		++r;
	return r;
}

uint32 ihypot2(int32 i,
	int32 j)
{
	int64		v, t;

	(void)muls64s32(v,i,i);
	(void)muls64s32(t,j,j);
	(void)int64add(v,t);
	return usqrt64(*(uint64*)&v);
}

uint32 uhypot2(uint32 i,
	uint32 j)
{
	uint64		v, t;

	(void)mulu64u32(v,i,i);
	(void)mulu64u32(t,j,j);
	(void)uint64add(v,t);
	return usqrt64(v);
}

uint32 ihypot3(int32 i,
	int32 j,
	int32 k)
{
	int64		v, t;

	(void)muls64s32(v,i,i);
	(void)muls64s32(t,j,j);
	(void)int64add(v,t);
	(void)muls64s32(t,k,k);
	(void)int64add(v,t);
	return usqrt64(*(uint64*)&v);
}

uint32 uhypot3(uint32 i,
	uint32 j,
	uint32 k)
{
	uint64		v, t;

	(void)mulu64u32(v,i,i);
	(void)mulu64u32(t,j,j);
	(void)uint64add(v,t);
	(void)mulu64u32(t,k,k);
	(void)uint64add(v,t);
	return usqrt64(v);
}
#endif

/* this LUT is a list of the fraction other than theta in the power series
  x**3/6 - ...
*/
static const uint32	ufsinlut[257] =
{
	0xA55DE731,0xA55DD2CA,0xA55D9596,0xA55D2F95,
	0xA55CA0C7,0xA55BE92C,0xA55B08C5,0xA559FF92,
	0xA558CD95,0xA55772CE,0xA555EF3E,0xA55442E7,
	0xA5526DC8,0xA5506FE5,0xA54E493D,0xA54BF9D3,
	0xA54981A8,0xA546E0BD,0xA5441716,0xA54124B4,
	0xA53E0998,0xA53AC5C5,0xA537593D,0xA533C404,
	0xA530061A,0xA52C1F83,0xA5281042,0xA523D85A,
	0xA51F77CD,0xA51AEE9E,0xA5163CD1,0xA5116269,
	0xA50C5F6A,0xA50733D7,0xA501DFB3,0xA4FC6303,
	0xA4F6BDCA,0xA4F0F00C,0xA4EAF9CD,0xA4E4DB12,
	0xA4DE93DF,0xA4D82438,0xA4D18C22,0xA4CACBA1,
	0xA4C3E2BA,0xA4BCD173,0xA4B597CF,0xA4AE35D5,
	0xA4A6AB88,0xA49EF8F0,0xA4971E10,0xA48F1AEF,
	0xA486EF92,0xA47E9BFF,0xA476203C,0xA46D7C4E,
	0xA464B03C,0xA45BBC0D,0xA4529FC5,0xA4495B6D,
	0xA43FEF09,0xA4365AA2,0xA42C9E3C,0xA422B9E1,
	0xA418AD96,0xA40E7962,0xA4041D4C,0xA3F9995D,
	0xA3EEED9B,0xA3E41A0D,0xA3D91EBC,0xA3CDFBAF,
	0xA3C2B0EE,0xA3B73E81,0xA3ABA470,0xA39FE2C2,
	0xA393F981,0xA387E8B4,0xA37BB065,0xA36F509B,
	0xA362C95F,0xA3561ABB,0xA34944B6,0xA33C475B,
	0xA32F22B1,0xA321D6C3,0xA314639A,0xA306C93E,
	0xA2F907BB,0xA2EB1F18,0xA2DD0F60,0xA2CED89E,
	0xA2C07AD9,0xA2B1F61E,0xA2A34A76,0xA29477EB,
	0xA2857E88,0xA2765E56,0xA2671762,0xA257A9B4,
	0xA2481559,0xA2385A5B,0xA22878C5,0xA21870A2,
	0xA20841FE,0xA1F7ECE3,0xA1E7715D,0xA1D6CF78,
	0xA1C6073F,0xA1B518BE,0xA1A40401,0xA192C913,
	0xA1816801,0xA16FE0D7,0xA15E33A1,0xA14C606B,
	0xA13A6742,0xA1284832,0xA1160348,0xA1039891,
	0xA0F1081A,0xA0DE51EF,0xA0CB761D,0xA0B874B2,
	0xA0A54DBB,0xA0920145,0xA07E8F5E,0xA06AF813,
	0xA0573B71,0xA0435987,0xA02F5263,0xA01B2611,
	0xA006D4A1,0x9FF25E21,0x9FDDC29E,0x9FC90226,
	0x9FB41CC9,0x9F9F1295,0x9F89E399,0x9F748FE2,
	0x9F5F1781,0x9F497A83,0x9F33B8F8,0x9F1DD2EF,
	0x9F07C877,0x9EF199A0,0x9EDB4678,0x9EC4CF0F,
	0x9EAE3375,0x9E9773BA,0x9E808FEC,0x9E69881C,
	0x9E525C5B,0x9E3B0CB7,0x9E239941,0x9E0C020A,
	0x9DF44721,0x9DDC6897,0x9DC4667C,0x9DAC40E2,
	0x9D93F7D8,0x9D7B8B6F,0x9D62FBB9,0x9D4A48C6,
	0x9D3172A7,0x9D18796E,0x9CFF5D2B,0x9CE61DF0,
	0x9CCCBBCE,0x9CB336D7,0x9C998F1C,0x9C7FC4B0,
	0x9C65D7A2,0x9C4BC807,0x9C3195EF,0x9C17416C,
	0x9BFCCA91,0x9BE2316F,0x9BC7761A,0x9BAC98A3,
	0x9B91991C,0x9B767799,0x9B5B342C,0x9B3FCEE7,
	0x9B2447DE,0x9B089F22,0x9AECD4C8,0x9AD0E8E3,
	0x9AB4DB84,0x9A98ACC0,0x9A7C5CAA,0x9A5FEB55,
	0x9A4358D4,0x9A26A53C,0x9A09D09F,0x99ECDB12,
	0x99CFC4A8,0x99B28D75,0x9995358D,0x9977BD03,
	0x995A23ED,0x993C6A5E,0x991E906B,0x99009627,
	0x98E27BA8,0x98C44101,0x98A5E648,0x98876B90,
	0x9868D0EF,0x984A167A,0x982B3C44,0x980C4264,
	0x97ED28EF,0x97CDEFF8,0x97AE9796,0x978F1FDE,
	0x976F88E5,0x974FD2C1,0x972FFD87,0x9710094C,
	0x96EFF626,0x96CFC42B,0x96AF7371,0x968F040E,
	0x966E7616,0x964DC9A2,0x962CFEC5,0x960C1597,
	0x95EB0E2E,0x95C9E8A0,0x95A8A503,0x9587436E,
	0x9565C3F8,0x954426B6,0x95226BC0,0x9500932B,
	0x94DE9D10,0x94BC8984,0x949A589F,0x94780A78,
	0x94559F25,0x943316BE,0x94107159,0x93EDAF0F,
	0x93CACFF6,0x93A7D425,0x9384BBB5,0x936186BC,
	0x933E3552,0x931AC790,0x92F73D8B,0x92D3975D,
	0x92AFD51C,0x928BF6E2,0x9267FCC5,0x9243E6DE,
	0x921FB544
};

ifract fastsin(ufract theta)
{
	ufract		sa, sb;
	unsigned int	t = theta, u;
/*
 theta' = theta{30}

 high two bits:
  00 (0.0pi): theta'' =  theta';    result =  sin(theta'')
  01 (0.5pi): theta'' =  pi-theta'; result =  sin(theta'')
  10 (1.0pi): theta'' =  theta';    result = -sin(theta'')
  11 (1.5pi): theta'' =  pi-theta'; result = -sin(theta'')
*/
	if ((t+=t)&(1L<<31))
		t =-t;
	/* get the two nearest values in the LUT of (theta mod pi/2) and interpolate */
	sb = ufsinlut[u=t>>23], sb-= ufract32mul(sb-ufsinlut[u+1],t<<9);
	/* theta''*(pi/2-theta''**2*...) */
	sa = 0xC90FDAA2-ufract32mul(ufract32mul(t,t)*2,sb);
	/* negative if >= pi */
	return !(theta&(1L<<31)) ? ufract32mul(sa,t):-ufract32mul(sa,t);
}

ufract ufractatan2(uint32 x,
	uint32 y)
{
	uint32		t, u, h;

	if (y<=(unsigned)x) /* y>x: return pi/2-atan2(y,x) */
	{
		u = 0;
		if (y==(unsigned)x) /* x=y: 0 ? pi/2:pi/4 -- pol(0,0) is actually undefined */
			{t = 1686629713+(1686629713&-!x); goto endo;}
	} else
		u = y, y = x, x = u, u = 0x80000000u;

	/* r = y/x */
	h = (y=ufract32div(y,x))&-1U<<26;
	{ /* r = atan(a)+atan((r-a)/(r*a+1)) */
		static const uint32		hibtable[64] =
		{
			0x00000000,0x03FFEAAB,0x07FF556F,0x0BFDC0C2,
			0x0FFAADDC,0x13F59F0E,0x17EE1826,0x1BE39EBE,
			0x1FD5BA9B,0x23C3F5F6,0x27ADDDD2,0x2B93023C,
			0x2F72F698,0x334D51D3,0x3721AEA5,0x3AEFABBE,
			0x3EB6EBF2,0x4277165F,0x462FD68C,0x49E0DC81,
			0x4D89DCDC,0x512A90DB,0x54C2B665,0x5852100C,
			0x5BD86508,0x5F55812E,0x62C934E5,0x66335515,
			0x6993BB0F,0x6CEA4477,0x7036D325,0x73794D0D,
			0x76B19C16,0x79DFADFC,0x7D03742D,0x801CE39E,
			0x832BF4A7,0x8630A2DB,0x892AECE0,0x8C1AD446,
			0x8F005D5F,0x91DB8F16,0x94AC72CA,0x97731420,
			0x9A2F80E6,0x9CE1C8E7,0x9F89FDC5,0xA22832DC,
			0xA4BC7D19,0xA746F2DE,0xA9C7ABDC,0xAC3EC0FC,
			0xAEAC4C39,0xB110688B,0xB36B31C9,0xB5BCC490,
			0xB8053E2C,0xBA44BC7E,0xBC7B5DEB,0xBEA94145,
			0xC0CE85B9,0xC2EB4ABB,0xC4FFAFFB,0xC70BD54D
		};

		t = hibtable[h>>26];
		y = ufract30div(y-h,(ufract32mul(y,h)+2>>2)+(1L<<30));
	}
	/* s = sqr(r) */
	h = ufract32mul(y,y);

	/* t+= r**1/1-r**3/3+r**5/5.. */
	t+= y;
	y = ufract32mul(y,h);
	t-= ufract32mul(y,((4294967295.0+3)/3));
	/* the table (above) is now large enough that what would follow underflows */
	t = (t+1)>>1;
endo: /* adjust for ufract format (value is currently in radians) */
	return (!u ? ufract32mul(t,2734261102U):u+-ufract32mul(t,2734261102U))+1>>1;
}

ufract ifractatan2(int32 x,
	int32 y)
{
/*
  n,p  2  .5pi   1 p,p
            |
       pi___|___ 0
            |
			I
  n,n  3  1.5pi  4 p,n
*/
	/* y<0: return 2*pi-atan2(x,-y) */
	/* x<0: return pi/2+atan2(y,-x) */
	if (y>=0)
		if (x>=0)
			return ufractatan2(x,y);
		else
			return (1L<<30)+ufractatan2(y,x=-x);
	else if (x>=0)
		return -ufractatan2(x,y=-y);
	else
		return 0xC0000000+-ufractatan2(y=-y,x=-x);
}

ifract fractasin(ifract x)
{
	uint64		y;
	ufract		a;

	if ((a=ABS(x))>(1<<30))
		return x<0 ? -1<<31:(1u<<31)-1;
	(void)mulu64u32(y,a,a);
	(void)uint64neg(y);
	(void)uint64addhl(y,1u<<60-32,0);
	return ifractatan2(usqrt64(y),x);
}

#if !defined(__POCC__) && !defined(envx86)
#undef cntlz32
int cntlz32(uint32 x)
{
	/* divide up logarithm (tree method) */
	if (x&0xFFFF0000u)
	{
		if (x&0xFF000000u)
			if (x&0xF0000000u)
				if (x&0xC0000000u)
					return (long)x>=0;
				else
					return !(x&0x20000000)+2;
			else if (x&0xC000000)
				return !(x&0x8000000)+4;
			else
				return !(x&0x2000000)+6;
		if (x&0xF00000)
			if (x&0xC00000)
				return !(x&0x800000)+8;
			else
				return !(x&0x200000)+10;
		if (x&0xC0000)
			return !(x&0x80000)+12;
		return !(x&0x20000)+14;
	}
	if (x&0xFF00)
		if (x&0xF000)
			if (x&0xC000)
				return !(x&0x8000)+16;
			else
				return !(x&0x2000)+18;
		else if (x&0xC00)
			return !(x&0x800)+20;
		else
			return !(x&0x200)+22;
	if (x&0xF0)
		if (x&0xC0)
			return !(x&0x80)+24;
		else
			return !(x&0x20)+26;
	if (x&12)
		return !(x&8)+28;
	return x&2 ? 30:32-x;
}
#endif
