/*
 interpreter.c
  Version 1.7.5

  Created: 2008.08.11

 1.7.5 (2013.05.20)
   added ALLOW_NAME_DOT and ALLOW_ESCNL.
 1.7.4 (2013.05.14)
   revised parse_textloop: custom parameter w. b func, removing 'a'.
   added parse_eval_literal and ALLOW_ELITERALS.
   fixed ALLOW_BIN_HI_LO.
 1.7.3 (2013.05.12)
   clarified _e_return.
   added several more valid/future error codes (not used here).
   added long  #[..]# comments and NO_EBREAKS.
   _e_no_identifier (no longer 'ID error') returned for missing ID in setv.
 1.7.2 (2013.05.03)
   parse_parameter returns error, error position.
 1.7.1 (2013.04.26)
   unary * or % treated as start of possible ID.
   added optional high/low < and > unary operations.
   '@' taken as part of ID if indirection function is n/a.
 1.7 (2013.04.03)
   end-pointer style for readanyint64 instead of length.
   fixed backslash in print_tchars.
 1.6.6 (2013.04.01)
   non-breaking space parsing now optional, and off by default.
   added parse_literal and readanyint64.
   renamed print_chars to print_tchars.
 1.6.5 (2013.03.28)
   fixed parse_textloop octal reading (3 excess and 'unset' bit set).
   added null ('Z') to finish single escapes for efficient string printing.
 1.6.4 (2013.02.24)
   readying opcode base for transition;
  double-bnot (!!b); explicit "load indirect"; "load value" now 'a=b'.
 1.6.3 (2013.02.20)
   floating-point-friendly _not operation does -1-b instead of ~b.
   24-bit packing for short storage for doubles instead of byte as used w. integers.
   fixed return operator bug.
   added another workaround for another POCC integer cast bug (erroneous code).
   fixed freadline bug: 'd = p' was missing in reallocation for termination.
   parse_number: added exp clipping and manual base-10 scaling for accuracy.
 1.6r2 (2013.02.17)
   fixed evaluate_chunk long double case, 'r' in ALLOW_RUNON_WSP (added undef).
 1.6 (2013.02.04)
   freadline may take an allocated ptr, reallocs with zero (frees) on empty EOF;
  this distinguishes empty lines from EOF.
   fixed some syntax error detection cases.

 See history.txt for the complete version history.

 Important Language Notes:
   Unary ops are done *after* run-on multiplication.
    Example: `~/2~/3` does 2*square root of 3, the square root of that (1.8612097182042)
   Unlike C, multiple interrogations (`?`) in a row are not handled, and will error.
    And because the initial 'else' value is taken, `a?(2):1?4:3` will always result
    in 4; first interrogation, `a?(2):1`, yields non-zero value.

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "interpreter.h"
#include "ftrig.h"

#ifdef env68k
 #pragma segment Interpreter
#endif

#ifndef ALLOW_RUNON_WSP
 #define ALLOW_RUNON_WSP	0
#endif
#ifndef STRICT_ESCAPES
 #define STRICT_ESCAPES		0
#endif

enum
{/* operator types */
 /* unary - all-type */
	_op_ldv,	/* (load) */
	_op_ldi,	/* @	(load indirect) */
	_op_srt,	/* ~/	[or]	CHR_RA */
	_op_abs,	/* + */
/*0x04*/_op_neg,	/* -	[or]	CHR_MN */
	_op_not,	/* ~	(compliment) */
	_op_bno,	/* !	(boolean not - no fp) */
	_op_dno,	/* double-not (optimized !!b) */
#ifdef ALLOW_BIN_HI_LO
	_op_blo,	/* <	(binary low byte) */
	_op_bhi,	/* >	(binary high byte) */
#endif
 /* dyadic logic/shift - no-fp */
/*0x08*/_op_bor,	/* || */
	_op_ban,	/* && */
	_op_ior,	/* | */
	_op_and,	/* & */
/*0x0C*/_op_eor,	/* ^ */
	_op_asr,	/* >>> */
	_op_shr,	/* >> */
	_op_shl,	/* << */
 /* dyadic math - all-type */
/*0x10*/_op_add,	/* + */
	_op_sub,	/* - */
	_op_mul,	/* * */
	_op_pow,	/* ** */
/*0x14*/_op_div,	/* / */
	_op_dvs,	/* //	[or]	CHR_DV */
	_op_mod,	/* % */
	_op_mds,	/* %% */
 /* dyadic compare - all-type */
/*0x18*/_op_clt,	/* <	(less than) */
	_op_cge,	/* >=	(greater or equal) */
	_op_cgt,	/* >	(greater than) */
	_op_cle,	/* <=	(less or equal) */
/*0x1C*/_op_cne,	/* !=	[or]	CHR_NE */
	_op_ceq,	/* = */
	_op_cni,	/* <>	(not identical) */
	_op_cid,	/* ==	(identical) */

/*0x20*/_op_stv,	/* ==>	(store) */
	_op_cml,	/* e.g., 2(x) */
	_op_apn,	/* .=>	(append) */
	_op_ape,	/* append empty */

	_op_ret,	/* counts as a terminator on the stack */
	_op_bra,	/* :	(branch)  [also] :> [label] */
	_op_brz,	/* ?	(branch on zero) */
	_op_bnz,	/* <?	(branch non-zero) */
	/* the dyadics are repeated in a 'set' context (e.g., `+=>` for 'add to') */
/* first to last unary -- must update if changed */
	_oc_una		= _op_ldv,
	_oc_unb		= _op_bor-1,
/* total operation mask (6 bits) */
	_oc_msk		= 0x3F,
	_oc_obj		= 0xFFFF	/* value (instead of a symbol) in parsing */
};

/* upper 8 bits are for small values, +/-0x7F */
typedef uint16		evalent;

#define ENTS_PER_EVAL		(sizeof(evalvalue)/sizeof(evalent))

/* EXT_EVAL: full value follows, lowest-significant first */
#if EVAL_FLOATING
 #define native_eval		double_t
 #define EXT_EVAL		0xFF
 typedef uint16		t_ceval;
#else
 #define native_eval		evalvalue
 #define EXT_EVAL		(int8)0x80
 typedef int16		t_ceval;
#endif

/* ---- macros */

#ifndef SKIP_NBSP
 #undef CHR_NB
#endif

#ifdef CHR_NB
 #define skipwsp(s,e)		do{uint8 c=*(s)++; \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0&&(c+=' '-CHR_NB)!=0) \
		{--(s); break;}}while(s<e)
 #define rskipwsp(s,f)		do{uint8 c=*--(s); \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0&&(c+=' '-CHR_NB)!=0) \
		{++(s); break;}}while(s>f)
#else
 #define skipwsp(s,e)		do{uint8 c=*(s)++; \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0) {--(s); break;}}while(s<e)
 #define rskipwsp(s,f)		do{uint8 c=*--(s); \
	if((c+=-'\t')!=0&&(c+='\t'-' ')!=0) {++(s); break;}}while(s>f)
#endif

#define skipdig(s,e)		do{uint8 c=*(s)++; \
	if ((c+=-'_')!=0 && (c+='_'-'0')>(uint8)('9'-'0')) \
	{--(s); break;}}while(s<e)

/* ---- globals */

#ifndef NO_TABLES
/* eol 0-3, space/tab/nbsp 9-11, alphanumerics 0x80-0xA3, 0xA4 is underscore */
static const char	upperctype[] =
{
	0x20,0x55,0x0C,0x0D,0x04,0x05,0x06,0x07,0x08,0x09,0x00,0x01,0x02,0x03,0x0E,0x0F,
	0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1A,0x1B,0x1C,0x1D,0x1E,0x1F,
	0x0A,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x29,0x2A,0x2B,0x2C,0x2D,0x2E,0x2F,
	0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x3A,0x3B,0x3C,0x3D,0x3E,0x3F,
	0x40,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,
	0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F,0xA0,0xA1,0xA2,0xA3,0x5B,0x5C,0x5D,0x5E,0xA4,
	0x60,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,
	0x99,0x9A,0x9B,0x9C,0x9D,0x9E,0x9F,0xA0,0xA1,0xA2,0xA3,0x7B,0x7C,0x7D,0x7E,0x7F,
#ifdef mac
	0x30,0x31,0x32,0x33,0x34,0x35,0x36,0xE7,0xCB,0xE5,0x30,0xCC,0x31,0x32,0x33,0xE9,
	0xE6,0xE8,0xEA,0xED,0xEB,0xEC,0x34,0xEE,0xF1,0xEF,0x35,0xCD,0xF2,0xF4,0xF3,0x36,
	0xCA,0x56,0x57,0x58,0x5F,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,
	0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xC6,0xB7,0xB8,0xB8,0xBA,0xBB,0xBC,0xBD,0xAE,0xAF,
	0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0x0B,0xCB,0xCC,0xCD,0xCE,0xCE,
	0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD9,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF,
	0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEA,0xEB,0xEC,0xED,0xEE,0xEF,
	0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFA,0xFB,0xFC,0xFD,0xFE,0xFF
#else /* windows-1252 */
	0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x41,0x42,0x43,0x44,0x45,0x46,
	0x46,0x47,0x48,0x49,0x4A,0x4B,0x4C,0x4D,0x4E,0x4F,0x50,0x51,0x43,0x52,0x53,0x54,
	0x0B,0x56,0x57,0x58,0x5F,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,0xAF,
	0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,0xBB,0xBC,0xBD,0xBE,0xBF,
	0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,
	0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,0xD8,0xD9,0xDA,0xDB,0xDC,0xDD,0xDE,0xDF,
	0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,
	0xF0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xF7,0xD8,0xD9,0xDA,0xDB,0xDC,0xFD,0xDE,0x9F
#endif
};

/* ---- system code */

int strcasecmp(const char *a,
	const char *b)
{
	const char	*t = upperctype;
	uint8		c, d;

	if (!a || !b)
		return IMIN(int);
	do
		if ((c=*a++)==0)
			return -*b;
	while((c=t[c])==(d=t[(uint8)*b++]));

	return c-d;
}

int strncasecmp(const char *a,
	const char *b,
	size_t z)
{
	const char	*t = upperctype;
	uint8		c, d;

	if (!a || !b)
		return IMIN(int);
	if ((signed)z<=0) /* unmapped space? */
		return 0;
	while((c=t[(uint8)*a++])==(d=t[(uint8)*b++]) && --z)
		;
	return c-d;
}

const char *parse_wordbreak(const char *s,
	const char *e,
	uint8 *h)
{
	if (!s)
		return 0;
	skipwsp(s,e);
	if (s==e)
		return s;
	if (h)
	{
		unsigned	c, v = 0;
		uint8		f = *s&0x3F;

		do
			if (upperctype[c=(uint8)*s++]>(char)0xA4)
				{--s; break;}
		while(v+=c,s<e);
		*h = f|v<<6;
	} else
		do
			if (upperctype[(uint8)*s++]>(char)0xA4)
				{--s; break;}
		while(s<e);
	return s;
}

const char *parse_rwordbreak(const char *s,
	const char *f)
{
	if (!s)
		return 0;
	rskipwsp(s,f);
	if (s==f)
		return s;
	do
		if (upperctype[(uint8)*--s]>(char)0xA4)
			{--s; break;}
	while(s>=f);

	return s;
}
#endif

/* extended standard */

long sreadline(const char *p,
	long z,
	char escape)
{
	const char	*s = p;
	uint8		c, u = 1;

	if (z<=0)
		return 0;
	do
	{
		c = *s++;
		if (escape && c==escape)
			u = 0;
		else
		{
			if ((c-=10)<4) /* LF/VT/FF/CR */
			{
				if (c==3 && *s==10) /* CR+LF = DOS line-ending */
					if (--z)
						++s;
					else
						++z;
				if (u)
					break; /* end of unescaped line */
			}
			u = 1;
		}
	} while(--z);

	return s-p;
}

char *freadline(FILE *f,
	uint32 *length,
	char *d,
	void *(reallocfunc)(void*, size_t),
	uint32 *lecount,
	int escape)
{
	void		*p;
	uint32		r, z, lec;
	uint8		e = 0;

	if (!reallocfunc)
		reallocfunc = realloc;
	lec = z = 0;
	if (lecount)
		lec = *lecount;
	if (!length)
		r =-1u;
	else if ((r=*length)==0)
		goto end;
	do
	{
		int		c;

		if ((c=fgetc(f))<0) //==EOF)
		{ /* empty on full EOF */
			if (d && !z)
				d = realloc(d,0);
			break;
		}
		if (escape && c==escape)
			e = 1;
		else
		{
			if ((unsigned)(c-10)<4u) /* LF/VT/FF/CR */
			{
				if (e>1)
				{
					e = 0;
					if (c==10) /* (DOS line-ending) */
						continue;
				} else if (c==13)
					e+= e;
				c = '\n', ++lec;
				if (!e) /* end of escaped line */
					break; /* last return not recorded */
			}
			e&=~1;
		}
		if (!(z&7)) /* add some extra bytes for fewer calls */
		{
			if ((p=reallocfunc(d,8+z))==0)
			{
			nix:	d = reallocfunc(d,0);
				goto end;
			}
			d = p;
		}
		d[z] = c;
		if (++z>0x7FFFFFEF)
			goto end; /* can't terminate */
	} while(--r);

	if (d)
	{ /* terminate */
		if (!(z&7))
		{
			if ((p=reallocfunc(d,1+z))==0)
				goto nix;
			d = p;
		}
		d[z] = 0;
	}
end:	if (length)
		*length = z;
	if (lecount)
		*lecount = lec;
	return d;
}

const char *readanyinteger(const char *s,
	long z,
	long int *v)
{
	const char	*e;
	uint32		r = 0, m = 0;
	uint8		c;

	if (!s || z<=0)
		return 0;
	e = z+s, c = *s++;
	if ((c+=-'-')==0)
	{
		if (s>=e)
			return 0;
		m =-1, c = *s++;
	} else if ((c+='-'-'+')==0)
	{
		if (s>=e)
			return 0;
		c = *s++;
	} else if ((uint8)((c+='+')-'0')>(uint8)9)
	{
#ifdef CHR_HX
		if (c==CHR_HX)
			goto hexadecimal;
#endif
		return 0;
	}
	if (c!='0')
	{
		goto decimal;
		do
		{
			c = *s++;
		decimal:
			c-= '0';
			if (c>(uint8)9)
				if ((c+='0'-'A')<=(uint8)('Z'-'A')
						|| (c+='A'-'a')<=(uint8)('z'-'a')
						|| (c+='a'-'_')==0)
					{s = 0; break;}
				else
					{--s; break;}
			r = r*10+c;
		} while(s<e);
		goto end;
	}

	if (s>=e)
		goto end;
	c = *s++;
	if (c!='x' && c!='X')
	{
		if (c=='o' || c=='O')
		{ /* octal (starting-zero-only not supported here) */
			do
			{
				c = *s++, c-= '0';
				if (c>(uint8)7)
					if (c<=(uint8)('9'-'7')
							|| (c+='0'-'A')<=(uint8)('Z'-'A')
							|| (c+='A'-'a')<=(uint8)('z'-'a')
							|| (c+='a'-'_')==0)
						{s = 0; break;}
					else
						{--s; break;}
				r = (r<<3)+c;
			} while(s<e);
			goto end;
		}
		if (c=='b' || c=='B')
		{ /* binary */
			do
			{
				c = *s++, c-= '0';
				if (c>(uint8)1)
					if (c<=(uint8)('9'-'1')
							|| (c+='0'-'A')<=(uint8)('Z'-'A')
							|| (c+='A'-'a')<=(uint8)('z'-'a')
							|| (c+='a'-'_')==0)
						{s = 0; break;}
					else
						{--s; break;}
				r+= r, r+= c;
			} while(s<e);
			goto end;
		}
		goto decimal;
	}
#ifdef CHR_HX
hexadecimal:
#endif
	if (s>=e) /* only starting character? */
		return 0;
	do
	{
		c = *s++, c-= '0';
		if (c<=(uint8)('9'-'0'))
			r = r*16+c;
		else if ((c+='0'-'A')<=(uint8)('Z'-'A')
				|| (c+='A'-'a')<=(uint8)('z'-'a'))
			if (c>(uint8)('F'-'A'))
				{s = 0; break;}
			else
				c+= 10, r = r*16+c;
		else if ((c+='a'-'_')==0)
			{s = 0; break;}
		else
			{--s; break;}
	} while(s<e);
end:	r+= m, r^= m;
	if (v)
		*v = r;
	return s;
}

const char *readanyint64(const char *s,
	const char *e,
	int64 *v)
{
	int64		r, t;
	uint32		m = 0;
	uint8		c;

	if (!s || s>e)
		return 0;
	if (s==e)
		return s;
	int64sethl(r,0,0);
	c = *s++;
	if ((c+=-'-')==0)
	{
		if (s>=e)
			return 0;
		m =-1, c = *s++;
	} else if ((c+='-'-'+')==0)
	{
		if (s>=e)
			return 0;
		c = *s++;
	} else if ((uint8)((c+='+')-'0')>(uint8)9)
	{
#ifdef CHR_HX
		if (c==CHR_HX)
			goto hexadecimal;
#endif
		return 0;
	}
	if (c!='0')
	{
		goto decimal;
		do
		{
			c = *s++;
		decimal:
			c-= '0';
			if (c>(uint8)9)
				if ((c+='0'-'A')<=(uint8)('Z'-'A')
						|| (c+='A'-'a')<=(uint8)('z'-'a')
						|| (c+='a'-'_')==0)
					{s = 0; break;}
				else
					{--s; break;}
			int64add(r,r), t = r;
			int64shl(t,2);
			int64add(r,t), int64addhl(r,0,c);
		} while(s<e);
		goto end;
	}

	if (s>=e)
		goto end;
	c = *s++;
	if (c!='x' && c!='X')
	{
		if (c=='o' || c=='O')
		{ /* octal (starting-zero-only not supported here) */
			do
			{
				c = *s++, c-= '0';
				if (c>(uint8)7)
					if (c<=(uint8)('9'-'7')
							|| (c+='0'-'A')<=(uint8)('Z'-'A')
							|| (c+='A'-'a')<=(uint8)('z'-'a')
							|| (c+='a'-'_')==0)
						{s = 0; break;}
					else
						{--s; break;}
				int64shl(r,3);
				(void)int64addhl(r,0,c);
			} while(s<e);
			goto end;
		}
		if (c=='b' || c=='B')
		{ /* binary */
			do
			{
				c = *s++, c-= '0';
				if (c>(uint8)1)
					if (c<=(uint8)('9'-'1')
							|| (c+='0'-'A')<=(uint8)('Z'-'A')
							|| (c+='A'-'a')<=(uint8)('z'-'a')
							|| (c+='a'-'_')==0)
						{s = 0; break;}
					else
						{--s; break;}
				int64add(r,r), (void)int64addhl(r,0,c);
			} while(s<e);
			goto end;
		}
		goto decimal;
	}
#ifdef CHR_HX
hexadecimal:
#endif
	if (s>=e) /* only starting character? */
		return 0;
	do
	{
		c = *s++, c-= '0';
		if (c<=(uint8)('9'-'0'))
		{
		hd:	int64shl(r,4);
			(void)int64addhl(r,0,c);
		} else if ((c+='0'-'A')<=(uint8)('Z'-'A')
				|| (c+='A'-'a')<=(uint8)('z'-'a'))
			if (c>(uint8)('F'-'A'))
				{s = 0; break;}
			else
			{
				c+= 10;
				goto hd;
			}
		else if ((c+='a'-'_')==0)
			{s = 0; break;}
		else
			{--s; break;}
	} while(s<e);
end:	(void)int64addhl(r,m,m);
	(void)int64eorhl(r,m,m);
	if (v)
		*v = r;
	return s;
}

int makebasestring(void *outstring,
	const char *basechars,
	long value,
	unsigned base,
	int minlength,
	char signchar)
{
	const char	*internalbasechars =
		"0123456789ABCDEFGHIJKLMNOPQRSTUVXYZ@#$`!%&()*+,./:;<=>?[\\]^{|}~";
	uint8		store[33]; /* enough space to do binary and sign */
	uint8		*s = sizeof(store)+store;
	unsigned	v = value;
	int		x = 0;

	if (!base)
		return 0;
	/* SC generated erroneous code if `value` wasn't used in the main loop, */
	/* optimizing it out entirely as `v`, nixing this conditional as a result */
	if ((signed)v<0 && signchar)
		v =-v;
	if (!basechars)
		basechars = internalbasechars;
	if (minlength<=0)
		minlength = 1;
	else if (minlength>sizeof(store))
		minlength = sizeof(store);
	do
	{
		unsigned	i = v;

		*--s = basechars[i-=(v/=base)*base];
	} while(++x<minlength || v);
	minlength = x;

	if (outstring)
	{
		uint8	*d = outstring;

		if (signchar && value<0)
			*d++ = signchar, ++minlength;
		/* copy to output--no func. calls, unaligned compatible */
		if (x&1)
			*d++ = *s++;
		if (x&2)
			*d++ = *s++, *d++ = *s++;
		if (x&=~3)
			do
				*d++ = *s++, *d++ = *s++, *d++ = *s++, *d++ = *s++;
			while(x-=4);
	}

	return minlength;
}

double_t scal10(double_t v,
	int e)
{
	static double_t	t[] = {10,100,1e4,1e8,1e16,1e32,1e64,1e128,1e256};
	const double_t	*a;
	int		m;

	if ((m=ABS(e))>500)
		m = 500;
	a = t;
	if (e>0)
	{
		if (!(~m&288))
			v*= 1e288, m&=~288;
		do
			if (m&1)
				v*= *a;
		while(++a,m>>=1);
	} else
	{
		if (!(~m&288))
			v/= 1e288, m&=~288;
		do
			if (m&1)
				v/= *a;
		while(++a,m>>=1);
	}
	return v;
}

const char *parse_number(const char *s,
	const char *e,
	double *_v,
	unsigned char r)
{
	double_t	fb; //can be a long double, coming from an integer
	double		v = 0;
	int		u = 0, a = 0, b = 10, m = 0, x = 0;
	uint8		c, n = 0;

	if (s>=e)
		return NULL;
	skipwsp(s,e);
#if CHR_MN!='-'
	if ((c=*s++)=='-' || c==(uint8)CHR_MN)
#else
	if ((c=*s++)=='-')
#endif
	{
		n = 0x80;
		if (s>=e)
			return 0;
		c = *s++;
	} else if (c=='+')
	{
		if (s>=e)
			return 0;
		c = *s++;
	} else if (c!='.' && (uint8)(c-'0')>(uint8)('9'-'0'))
		return 0; /* first must be a digit */

	if (c=='0' && s<e)
	{
		if ((c=*s-'X')==0 || (c+='X'-'x')==0)
			b = 16, u = 3, ++s, r = 0; /* fractionless hexadecimal */
		else if ((c+='x'-'h')==0)
			b = 16, u = 3, ++s; /* hexadecimal w/ fraction */
		else if ((c+='h'-'O')==0 || (c+='O'-'o')==0)
			b = 8, ++s; /* octal (starting-zero-only not supported) */
		else if ((c+='o'-'b')==0)
			b = 2, ++s; /* binary */
		else if ((c+='b'-'d')==0)
			b = 10, r = 0, ++s; /* fractionless decimal */
		else if ((c+='d'-'z')==0)
			b = 36, u = 3, ++s; /* base 36 (all alphanumeric) */
		else
			u|= 0x8000; /* decimal, zero parsed */
		if (s>=e)
			return 0; /* must follow digits */
		c = *s++;
	}
	fb = b;
	goto start;
	do
	{
		c = *s++;
	start:	if ((c+=-'0')<=(uint8)('9'-'0')) /* 0-9 */
		{
			if (c>=(uint8)b)
				return 0; /* digit invalid for this base number */
			u|= 0x8000;
			if (a>0) /* this could easily overflow without clipping */
				x = x>99999999 ? 1000000000:10*x+c;
			else
				v = fb*v+c, m+= a;
		} else if ((c+='0'-'A')<='Z'-'A' || (c+='A'-'a')<='z'-'a' && (uint8)u>2)
		{
			if ((c+=10)>=(uint8)b)
				return 0; /* digit invalid for this base number */
			v = b*v+c, m+= a, u|= 0x8000;
			continue;
		} else if ((c+='A'-'E')==0) /* Ee */
		{ /* can't have more than one `E' */
			if (a==1)
				return 0;
			a = 1, u&=~0x8000;
		} else if ((c=s[-1]-'.')==0) /* . */
		{ /* can't have multiple points or for integer or notation */
			if (a || !r)
				return 0;
			a =-1;
#if CHR_MN!='-'
		} else if ((c+='.'-'-')==0 || c==(uint8)(CHR_MN-'-'))
#else
		} else if ((c+='.'-'-')==0)
#endif
		{ /* only one '-' valid here for notation; cannot come after `E+' */
			if (a<=0 || u&0x8001)
				goto back;
			u = 1;
		} else if ((c+='-'-'+')==0) /* + */
		{ /* only one '+' valid here for notation; cannot come after `E-' */
			if (a<=0 || u&0x80FF)
				goto back;
			u = 2;
		} else if ((c+='+'-'_')==0)
			return 0; /* can't directly follow an underscore */
		else /* everything else ends */
		{
		back:	--s;
			break;
		}
	} while(s<e);

	if (!(u&0x8000))
		return 0; /* no digit chars after start or notation */
	if (_v)
	{
		if ((uint8)u==1)
			m-= x;
		else
			m+= x;
		((int8*)&v)[!BIG_ENDIAN*(sizeof(double)-1)]^= n;
		*_v = scal10(v,m);
	}
	return s;
}

/* this requires that the characters have already been processed */
/* (use parse_textloop) */
int parse_literal(const char *s,
	int64 *r,
	long z)
{
	const char	*e;
	int64		v;

	if (z>8)
		return -1;
	e = s+z;
	int64sethl(v,0,0);
	while(s<e)
	{
		int64shl(v,8);
		(void)int64orhl(v,0,*s++);
	}
	if (r)
		*r = v;
	return z<3 ? z:z<5 ? 4:8;
}

static void getl(void *u,
	char v)
{
	**(char**)u = v, ++*(char**)u;
}

const char *parse_eval_literal(const char *str,
	const char *e,
	evalvalue *v)
{
	char		b[10], *p = b;
	const char	*s;
	int64		i;
	long		z = 8;

	s = parse_textloop(str,e,&z,NULL,NULL,&p,getl);
	parse_literal(b,&i,z);
#if EVAL_FLOATING
	*v = int64tox(i);
#else
	*v = int64tol(i);
#endif
	return *s==*str ? s+1:NULL;
}

static int makeUTF8(uint32 v,
	uint8 *d)
{
	int		z = 0, i = 0;

	if (v<0x80)
		z = 1;
	else if (v<0x800)
		z = 2;
	else if (v<0x10000)
		z = 3;
	else if (v<0x200000)
		z = 4;
	else /* 'invalid' replacement char */
		z = 2, v = 0xFFFD;
	d[i=z] = 0;
	while(--i)
		d[i] = (v&0xBF)|0x80, v>>= 6;
	*d = v|(int8)(-0x200>>z)>>1;
	return z;
}

const char *parse_textloop(const char *t,
	const char *e,
	long *_z,
	uint32 *_t,
	char **r,
	void *u,
	void (*b)(void *u,
		char v))
{
	static uint8	ub[6];
	uint8		*us;
	const char	*s;
	char		*p = 0;
	uint32		m;
	long		z = 0, w, vc;
	uint8		c, q, l;

	if ((s=t)==0 || s>=e)
		return s;
	us = NULL, z = m = vc = w = 0, l = 0, q = *s;
	if (q!='"' && q!='\'')
		return s;
	if (r)
		p = *r;
	while(++s<e)
	{
		c = *s;
		if (w)
			if (!--vc)
				goto form;
			else if ((uint8)(c-'0')>(uint8)9)
			{
				if (w>3)
				{
					if ((c+=-'A')<=(uint8)('F'-'A')
							|| (c+='A'-'a')<=(uint8)('f'-'a'))
						{c+= 10, m = m<<4|c, w&=63; continue;}
					if (w&0x80) /* no digits */
						goto bad;
				}
			form:	if (us)
					{makeUTF8(m,us); c = *us++, m = 0;}
				else
					c = m;
				if (m>255) /* no ind. chars out of range */
					goto bad;
				m = w = 0, --s;
			} else if ((c-='0')>>w)
				goto bad;
			else
				{m = m<<w|c, w&=63; continue;}
		else if (c==q)
			break;
		else if (c=='\\')
		{
			if (++s>=e)
			{
			bad:	z|=-1u<<31;
				goto end;
			}
			switch(c=*s)
			{
#if STRICT_ESCAPES
			case 'r': c = CHR_NL^7; break; /* carriage return */
#else
			case 'Z': c = 0; break; /* null */
			case 'h': c = 1; break; /* heading start */
			case 'S': c = 2; break; /* text start */
			case 'e': c = 3; break; /* text end/enter */
			case 'E': c = 4; break; /* end transmission */
			case 'q': c = 5; break; /* enquiry */
			case 'k': c = 6; break; /* acknowledge */
			case 'l': c = 10; break; /* linefeed */
			case 'v': c = 11; break; /* vertical tab */
			case 'r': c = 13; break; /* carriage return */
			case 'o': c = 14; break; /* shift out */
			case 'i': c = 15; break; /* shift in */
			case 'd': c = 16; break; /* data link esc. */
			case 'N': c = 17; break; /* dev.ctl.1-XN */
			case 'A': c = 18; break; /* device control 2 */
			case 'O': c = 19; break; /* dev.ctl.3-XO */
			case 'D': c = 20; break; /* dev.ctl.4-stop */
			case 'z': c = 21; break; /* negative ack./zero */
			case 'j': c = 22; break; /* sync. idle/join */
			case 'w': c = 23; break; /* text block end/word */
			case 'c': c = 24; break; /* cancel */
			case 'm': c = 25; break; /* medium end */
			case 's': c = 26; break; /* substitute */
			case 'y': c = 27; break; /* escape/yield */
			case 'F': c = 28; break; /* file separator */
			case 'G': c = 29; break; /* group separator */
			case 'R': c = 30; break; /* record separator */
			case 'U': c = 31; break; /* unit separator */
#endif
			case 'a': c = 7; break; /* bell/alert */
			case 'b': c = 8; break; /* backspace */
			case 't': c = 9; break; /* horiz. tab */
			case 'f': c = 12; break; /* formfeed */
			case 'n': c = CHR_NL; break; /* new line */
			case 'p': /* Pascal str */
				if (!z) /* only on starting byte */
					l = 1;
				break;
			case 'u': /* unicode (UTF-8) character */
				us = ub, vc = 3;
			case 'x': /* hexadecimal */
				w = 0x84;
			bc:	vc+= 3;
				continue;

			default:
				if ((uint8)(c-'0')<=(uint8)7) /* octal */
					{c&= 7, m = c, w = 0x03; goto bc;}
				else if ((uint8)(c-10)<=(uint8)3) /* line-ending */
					continue; /* escaped here = ignored */
			}
		}
	append:	if (b)
			b(u,c);
		if (r)
		{
			if (!(z&7))
				if ((p=realloc(p,8+z))==0)
				{
					*r = realloc(*r,0);
					z =-1<<31;
					goto end;
				} else
					*r = p;
			p[z] = c;
		}
		if (++z>0x7FFFFFF0)
			break;
		if (us)
			if ((c=*us++)!=0)
				goto append;
			else
				us = NULL;
	}
	if (l)
	{
		if (z>255+1)
			z = 256; /* clip overflow */
		*p = z-1; /* length byte */
		if (_t)
			*_t = _t_uint8; /* _t_pstr */
	} else if (_t)
		*_t = _t_int8; /* p[z-1] ? _t_uint8:_t_int8 */
end:	if (*_z)
		*_z = z;
	return s;
}

uint32 print_tchars(FILE *f,
	const char *s,
	long z,
	char q)
{
	const char	*t = "ZhSeEqkabtlvfr";
	long		r;
	uint8		c;

	if (!f || !s || z<=0)
		return 0;
	r = 0;
	do
		if ((c=*s++)<14)
			r+= fputc('\\',f), r+= fputc(t[c],f);
		else if (c==q || c=='\\')
			r+= fputc('\\',f), r+= fputc(c,f);
		else
			r+= fputc(c,f);
	while(--z);

	return r;
}

const char *parse_skipwsp(const char *s,
	const char *e)
{
	if (!s)
		return 0;
	skipwsp(s,e);

	return s;
}

const char *parse_rskipwsp(const char *s,
	const char *f)
{
	if (!s)
		return 0;
	rskipwsp(s,f);

	return s;
}

/* ---- evaluation */

/* single-char op. (e.g., `/`) is unsigned when <=4 bytes */
#if EVAL_FLOATING
 #define evalsing		evalvalue
#else
 #define evalsing		unsigned
#endif

static evalvalue esqrt(evalvalue a)
{
#if EVAL_FLOATING
	return sqrt(a);
#else
	return usqrt(a);
#endif
}

static evalvalue epow(evalvalue a,
	evalvalue b)
{
#if EVAL_FLOATING
	if (a<0)
	{ /* test for odd root of negative (valid) */
		evalvalue	c = 1/b;

		if (!(c-(int)c))
			return -pow(-a,b);
	}
	return pow(a,b);
#else
	if (b<=0)
		return !a ? EVAL_HUGEVAL:!b ? 1:++b ? 0:1/a;
	else if (a==1 || !--b)
		return a;
	else if (b>31)
		return !a ? 0:++a ? EVAL_HUGEVAL:(~(long)b&1)*2-1;
	else
	{
		int64		v;
		evalvalue	m = a;

		do
		{
			(void)muls64s32(v,a,m);
			if ((long)((a=int64lo(v))>>31)!=int64hi(v))
				return EVAL_HUGEVAL; /* overflow */
		} while(--b);
	}
	return a;
#endif
}

/* packing and unpacking functions for evalvalues */
static evalent *unpack_eval(evalent *a,
	t_ceval o,
	evalvalue *v)
{
#if EVAL_FLOATING
	if (o<(EXT_EVAL<<8))
	{
		((int32*)v)[!!BIG_ENDIAN] = 0;
		((int32*)v)[!BIG_ENDIAN] = (((uint32)(o&0xFF00)<<8)|*--a)<<8;
		return a;
	}
	((int16*)v)[!!BIG_ENDIAN*3] = *(a-=4);
	((int16*)v)[!!BIG_ENDIAN+1] = a[1];
	((int16*)v)[!BIG_ENDIAN+1] = a[2];
	((int16*)v)[!BIG_ENDIAN*3] = a[3];
#else
	if (o>(EXT_EVAL<<8)+0xFF)
	{
		*v = o>>8;
		return a;
	}
	((int16*)v)[!!BIG_ENDIAN] = *(a-=2);
	((int16*)v)[!BIG_ENDIAN] = a[1];
#endif
	return a;
}

static evalent *pack_eval(evalent *a,
	evalent *h,
	const evalvalue *v)
{
	uint32		b = ((int32*)v)[BIG_ENDIAN ? 0:sizeof(evalvalue)/8];

	*h = (uint8)EXT_EVAL<<8;
#if EVAL_FLOATING
	if (b<(EXT_EVAL<<24) && !(((int32*)v)[!!BIG_ENDIAN*3]|(b&0xFF)))
	{ /* short-encoding: high 24 bits of double, zero low-fraction */
		*h = ((int16*)v)[!BIG_ENDIAN*3]&0xFF00;
		a[0] = b>>8;
		return a+1;
	}
	a[0] = ((int16*)v)[!!BIG_ENDIAN*3];
	a[1] = ((int16*)v)[!!BIG_ENDIAN+1];
	a[2] = b;
	a[3] = ((int16*)v)[!BIG_ENDIAN*3];
#else
	if (b-(EXT_EVAL+1)<0xFF)
		{*h = b<<8; return a;}
	a[0] = b;
	a[1] = ((int16*)v)[!BIG_ENDIAN];
#endif
	return a+ENTS_PER_EVAL;
}

/*
 Entry bits:
  #7: interrogation (`?`), or already packed (doesn't apply here)
  #6: with interrogation, else (`:`)
*/

/* evaluate_chunk evaluates all operations until after a normal dyadic
 or none (run-on multiplication is not a normal dyadic)

 returns next stack ptr
*/
static evalent *evaluate_chunk(evalent o,
	const evalent *x,
	evalent *s,
	evalvalue *v,
	evalvalue (*geti)(evalvalue))
{
	evalent		c;

	while((c=o&_oc_msk)!=_op_ret)
	{
		evalvalue	a;
		native_eval	b;
		int32		t;

		if ((evalent)(c-_oc_una)<=_oc_unb-_oc_una)
			a = 0;
		else
			s = unpack_eval(s,o,&a);
		if (v)
		{
			b = *v;
			switch(c)
			{
			#include "evaluate.h"

			default: a = 0;
			}
			*v = a;
		}
		if (s<=x || (c>_oc_unb && c!=_op_cml))
			break;
		o = *--s;
	}
end:	return s;
}

/* ---- general line processing */

#define EVAL_STACKSIZE		(ENTS_PER_EVAL*65+1)

#define SET_OP(c)		(void)(o=c,h=(evalent)(EXT_EVAL<<8))

long evaluate_line(const char *str,
	const char *e,
	evalvalue *v,
	int *_err,
#ifdef ALLOW_LONG_COMMENTS
	uint8 *_com,
#endif
	const char *(*getv)(const char *s, const char *e, evalvalue*, void *u),
	const char *(*setv)(const char *s, const char *e, evalvalue, void *u),
	void *u,
	evalvalue (*geti)(evalvalue))
{
	evalent		xstack[EVAL_STACKSIZE];
	evalent		*a = xstack;
	const char	*p, *s = str;
	evalparse	i;
	int		err = 0;
#ifdef ALLOW_LONG_COMMENTS
	uint8		com = _com ? *_com:0;
#endif
	evalent		c; /* operator to be set */
	evalent		o = _op_ret; /* previous op. */
	evalent		h = 0; /* high opcode bits */
	uint8		q = 0; /* interrogation state; 0x80=then, 0xC0=else */
#if ALLOW_RUNON_WSP
 #define r			1
#else
	uint8		r = 0; /* when run-on mult. is allowed */
#endif
	while(s<e)
	{
#ifdef ALLOW_LONG_COMMENTS
		if (com)
		{
			if (*s++==']' && s<e && *s=='#')
				++s, com = 0;
			continue;
		}
#endif
		switch((uint8)*s) /* try not to get stuck */
		{
		case ' ':
		whitespace:
			skipwsp(s,e);
#if !ALLOW_RUNON_WSP
			r = 0;
#endif
			break;
		case '#':
			++s; /* ignore first '#' for numeric */
			if (s>=e || *s=='#') /* end of line/comment */
				{--s; goto end;}
#ifdef ALLOW_LONG_COMMENTS
			if (*s=='[')
				{com = 1, ++s; continue;}
#endif
#ifndef ALLOW_NAME_DOT
		case '.':
			if (s>=e)
				{--s; goto bad;} /* period by itself */
#endif
#ifdef CHR_HX
		case CHR_HX:
#endif
		CASE_NUMERIC
			/* value w/o parentheses */
			if (o==_oc_obj || (p=parse_eval(s,e,&i))==0)
				goto bad;
			else
				s = p;
			if (!(q&0x80))
				*v = i;
#if !EVAL_FLOATING
			if (*s=='.')
			{ /* skip unused fraction */
				++s;
				skipdig(s,e);
			}
#endif
		runon:	//++n; /* increment counter */
#if ALLOW_RUNON_WSP
			skipwsp(s,e);
#endif
			if (s<e)
				if (*s=='(')
					goto paren;
				else if (*s=='@' || *s=='~'
#ifdef CHR_RA
						|| *s==CHR_RA
#endif
						)
				{ /* run-on mul. */
					if (a>=xstack+EVAL_STACKSIZE-ENTS_PER_EVAL-1)
						{err = _e_stack_error; goto end;} /* stack overflow */
					*a++ = o|h, o = _op_cml, a = pack_eval(a,&h,v);
					break;
				}
			if ((a=evaluate_chunk(o|h,xstack,a,q&0x80 ? 0:v,geti))<xstack)
				{err = _e_stack_error; goto end;}
			o = _oc_obj;
#if !ALLOW_RUNON_WSP
			r = 1;
#endif
			break;
		case '!':
			if (o==_op_bno)
				{o = _op_dno; ++s; break;}
			c = _op_bno;
			goto unary;
		case '~': c = _op_not;
		unary:	++s;
			skipwsp(s,e);
			if (c==_op_bno && *s=='=') /* != */
				{c = _op_cne; goto dyadic;}
			else if (c==_op_not && *s=='/') /* ~/ */
				c = _op_srt, ++s;
			if (o==_oc_obj)
				goto bad;
			if (a>=xstack+EVAL_STACKSIZE-1)
				{err = _e_stack_error; goto end;} /* stack overflow */
			*a++ = o|h;
			o = c;
			break;
		case '+': c = _op_abs;
			if (o==_op_abs)
				{++s; break;} /* 2013.02.24 */
			if (o==_oc_obj)
				{c = _op_add; goto dyadic;}
			goto unary;
		case '-':
			if (o==_oc_obj)
				{c = _op_sub; goto dyadic;}
			if (o==_op_neg)
				{o = _op_ldv; ++s; break;}
			c = _op_neg;
			goto unary;
		case '%': c = _op_mod;
			if (o==_op_mod) /* %% */
				c = _op_mds|0x80, o = _oc_obj;
			else if (o!=_oc_obj && getv)
				goto gid;
			goto dyadic;
		case '&': c = _op_and;
			if (o==_op_and) /* && */
				c = _op_ban|0x80, o = _oc_obj;
			goto dyadic;
		case '*':
			c = _op_mul;
			if (o==_op_mul) /* ** */
				c = _op_pow|0x80, o = _oc_obj;
			else if (o!=_oc_obj && getv)
				goto gid;
			goto dyadic;
		case '/': c = _op_div;
			if (o==_op_div) /* // */
				c = _op_dvs|0x80, o = _oc_obj;
			goto dyadic;
		case '<': c = _op_clt;
			if (o==_op_clt) /* << */
				c = _op_shl|0x80, o = _oc_obj;
#ifdef ALLOW_BIN_HI_LO
			else if (o!=_oc_obj)
				{c = _op_blo; goto unary;}
#endif
			goto dyadic;
		case '=': c = _op_ceq;
			if (o==_op_ceq) /* == */
				c = _op_cid|0x80, o = _oc_obj;
			else if (o==_op_clt) /* <= */
				c = _op_cle|0x80, o = _oc_obj;
			else if (o==_op_cgt) /* >= */
				c = _op_cge|0x80, o = _oc_obj;
			goto dyadic;
		case '>': c = _op_cgt;
			if (o==_op_clt) /* <> */
				c = _op_cni|0x80, o = _oc_obj;
			else if (o==_op_shr) /* >>> */
				c = _op_asr|0x80, o = _oc_obj;
			else if (o==_op_cgt) /* >> */
				c = _op_shr|0x80, o = _oc_obj;
			else if (o==_op_cid) /* ==> */
			{
				o = _op_stv, ++s;
				break;
			}
#ifndef NO_EBREAKS
			else if (s+1<e && s[1]==';') /* '>;' */
			{
				if (q&0x80) /* this part not being executed */
				{
					/*o = _oc_obj;*/
					s+= 2;
					break;
				}
				if (_err)
					*_err = 0;
 #ifdef ALLOW_LONG_COMMENTS
				if (_com)
					*_com = com;
 #endif
				return _e_return;
			}
#endif
#ifdef ALLOW_BIN_HI_LO
			else if (o!=_oc_obj)
				{c = _op_bhi; goto unary;}
#endif
			goto dyadic;
		case '|': c = _op_ior;
			if (o==_op_ior) /* || */
				c = _op_bor|0x80, o = _oc_obj;
			goto dyadic;
		case '^': c = _op_eor;
		dyadic:	if (o!=_oc_obj)
			{ /* dyadic after dyadic or unary */
				if (o!=_op_ret)
					--s; /* two in a row; discount 2nd */
				goto bad;
			}
			if (a>=xstack+EVAL_STACKSIZE-ENTS_PER_EVAL)
				{err = _e_stack_error; goto end;} /* stack overflow */
			if (!(c&0x80)) /* if not already packed */
				a = pack_eval(a,&h,v);
			o = c&_oc_msk, ++s;
			break;
#ifndef NO_EBREAKS
		case ';': /* break */
			++s; /* incomplete expression tests */
			if (o==_op_ret)
			{
				if (q&0xC0)
					goto end;
			} else if ((int8)o>0 || (o!=_oc_obj
					&& (a=unpack_eval(a,h,&i))!=xstack))
				goto end;
			o = _op_ret;
			break;
#endif
		case ',': /* parameter */
			goto end;
		case '(':
			if (o==_oc_obj)
			{ /* run-on multiply */
				if (!r)
				{ /* concat not implemented */
					err = _e_param_separation;
					goto end;
				}
			paren:	if (a>=xstack+EVAL_STACKSIZE-ENTS_PER_EVAL-2)
					{err = _e_stack_error; goto end;} /* stack overflow */
				if (o==_oc_obj)
					o = _op_ret;
				*a++ = o|h, o = _op_cml, a = pack_eval(a,&h,v);
			} else if (a>=xstack+EVAL_STACKSIZE-1)
				{err = _e_stack_error; goto end;} /* stack overflow */
			o&= _oc_msk, o|= q; /* save interrogation state */
			*a++ = o|h;
			SET_OP(_op_ret);
			if ((q&=0x80)==0)
				*v = 0; /* init. interrogation result */
			++s;
			break;
		case ')':
			if (o!=_oc_obj)
				goto end;
			if (a<=xstack)
				{err = _e_stack_error; goto end;} /* after op., empty or underflow */
			o = *--a;
			q = o&~_oc_msk; /* restore interrogation state */
			if ((a=evaluate_chunk(o,xstack,a,q&0x80 ? 0:v,geti))<xstack)
				goto bad;
			SET_OP(_oc_obj);
#if !ALLOW_RUNON_WSP
			r = 1;
#endif
			++s;
			break;
		case '@':
			if (geti)
				{c = _op_ldi; goto unary;}
#ifdef ALLOW_NAME_DOT
		case '.':
#endif
		case '_':
		CASE_ALPHABETIC
			if (o==_op_stv)
			{ /* set value for identifier(s); 2011.06.11 */
				if (!setv)
					{err = _e_handler_error; goto end;}
 				if ((a=unpack_eval(a,h,&i))<xstack)
					{err = _e_stack_error; goto end;}
				if ((p=setv(s,e,i,u))==0)
					{err = _e_no_identifier; goto end;}
			} else
			{ /* get value for identifier(s) */
				if (!getv) /* no routine, two in a row or no value */
					{err = _e_handler_error; goto end;}
			gid:	if (o==_oc_obj)
					goto bad;
				if ((p=getv(s,e,v,u))==0)
					{err = *s=='*' || *s=='%' ? -1:_e_no_identifier; goto end;}
				s = p;
				goto runon;
			}
			s = p;//, ++n; /* update, increment counter */
			o = _oc_obj;
			break;
		case '?':
			if (o!=_oc_obj || q&0x40) /* oper after oper or interrogation open */
				{err = _e_open_interrogator; goto bad;}
			/* eval. 'else' range only if 0; otherwise 'then' only */
			q = *v ? 0x40:0xC0;
			*v = 0; /* value parsed; clear */
			o = _op_ret;
			++s;
			break;
		case ':':
			if (o!=_oc_obj /* oper after oper, or unimplemented */
					|| !(q&0x40))
				{err = _e_incomplete_error; goto bad;}
			q^= 0xC0; /* clears bit #6 */
			o = _op_ret;
			++s;
			break;
#ifdef ALLOW_ELITERALS
		case 047:
			if (o==_oc_obj)
				goto bad;
			if ((s=parse_eval_literal(s,e,v))!=NULL)
				goto runon;
			else
#endif
		case 042:
			goto bad;
#ifdef ALLOW_ESCNL
		case '\\':
			if (++s>=e || *s-10u>3u)
				goto bad;
			++s;
			break;
#endif
		default:
#ifdef CHR_NE
			if ((uint8)*s==(uint8)CHR_NE)
				{c = _op_cne; goto dyadic;}
#endif
#ifdef CHR_LE
			if ((uint8)*s==(uint8)CHR_LE)
				{c = _op_cle; goto dyadic;}
#endif
#ifdef CHR_GE
			if ((uint8)*s==(uint8)CHR_GE)
				{c = _op_cge; goto dyadic;}
#endif
#ifdef CHR_RA
			if ((uint8)*s==(uint8)CHR_RA)
				{c = _op_srt; goto unary;}
#endif
#ifdef CHR_SL
			if ((uint8)*s==(uint8)CHR_SL)
				{c = _op_shl; goto dyadic;}
#endif
#ifdef CHR_SR
			if ((uint8)*s==(uint8)CHR_SR)
				{c = _op_shr; goto dyadic;}
#endif
#if CHR_MN!='-'
			if ((uint8)*s==(uint8)CHR_MN)
				{c = _op_neg; goto unary;}
#endif
#ifdef CHR_DV
			if ((uint8)*s==(uint8)CHR_DV)
				{c = _op_dvs; goto dyadic;}
#endif
#if defined(CHR_NB)
			if ((uint8)*s==(uint8)CHR_NB)
				goto whitespace;
#endif
			if ((uint8)*s==(uint8)'\t')
				goto whitespace;
		bad:	if (!err)
				err = _e_general_error;
			goto end;
		}
	}
end:	if (((int8)o>=0 && o!=_op_ret || a>xstack) && !err)
		err = o==_op_stv ? _e_no_identifier:_e_incomplete_error;
	if (_err)
		*_err = err;
#ifdef ALLOW_LONG_COMMENTS
	if (_com)
		*_com = com;
#endif
	return s-str;
}

#undef r

static const char *quotedtext_scan(const char *s,
	const char *e,
	uint8 v)
{
	if (e<=s)
		return 0;
	do
	{
		uint8	c = *s++;

		if (c=='\\') /* escape; skip next char */
			++s;
		else if (c==v)
			return s;
	} while(s<e);

	return 0; /* no closing quote */
}

const char *parse_parameter(const char *text,
	long length,
	char *outstring,
	int *outerr,
	const char *(*getv)(const char *s, const char *e, evalvalue*, void *u),
	const char *(*setv)(const char *s, const char *e, evalvalue, void *u),
	void *ud,
	evalvalue (*geti)(evalvalue))
{
	const char	*s, *e;
	evalvalue	v = 0;
	long		r;
	int		err;

	if (outstring)
		*outstring = '\0';

	if ((s=text)==0)
		return 0;
	if (length<=0)
		goto blank;
	e = length+s;
	skipwsp(s,e);

	if (s==e)
	{
	blank:	if (outstring)
			*outstring = ' ';
		goto end;
	}

	if (*s=='\042') /* text - double quote */
		{s = quotedtext_scan(++s,e,'\042'); goto end;}
#ifndef ALLOW_ELITERALS
	else if (*s=='\047') /* text - single quote */
		{s = quotedtext_scan(++s,e,'\047'); goto end;}
#endif
	r = evaluate_line(text=s,e,&v,&err,
#ifdef ALLOW_LONG_COMMENTS
		NULL,
#endif
		getv,setv,ud,geti);
	if (outerr)
		*outerr = err;
	if (r!=_e_return)
		s+= r;
	if (err<0)
		goto end;
	if (outstring)
#if EVAL_FLOATING
		if (v-(long)v)
			sprintf(outstring,"%.16g",
				v);
		else
#endif
			outstring[makebasestring(outstring,0,v,10,0,'-')] = 0;
end:	return s;
}
